# -*- coding: utf-8; -*-

from unittest import TestCase
from unittest.mock import MagicMock, patch

import sqlalchemy as sa

import colander
import deform
from pyramid import testing

from wuttjamaican.conf import WuttaConfig
from wuttaweb.forms import base, widgets
from wuttaweb import helpers
from wuttaweb.grids import Grid


class TestForm(TestCase):

    def setUp(self):
        self.config = WuttaConfig(defaults={
            'wutta.web.menus.handler_spec': 'tests.util:NullMenuHandler',
        })
        self.app = self.config.get_app()
        self.request = testing.DummyRequest(wutta_config=self.config, use_oruga=False)

        self.pyramid_config = testing.setUp(request=self.request, settings={
            'mako.directories': ['wuttaweb:templates'],
            'pyramid_deform.template_search_path': 'wuttaweb:templates/deform',
        })

    def tearDown(self):
        testing.tearDown()

    def make_form(self, **kwargs):
        return base.Form(self.request, **kwargs)

    def make_schema(self):
        schema = colander.Schema(children=[
            colander.SchemaNode(colander.String(),
                                name='foo'),
            colander.SchemaNode(colander.String(),
                                name='bar'),
        ])
        return schema

    def test_init_with_none(self):
        form = self.make_form()
        self.assertEqual(form.fields, [])

    def test_init_with_fields(self):
        form = self.make_form(fields=['foo', 'bar'])
        self.assertEqual(form.fields, ['foo', 'bar'])

    def test_init_with_schema(self):
        schema = self.make_schema()
        form = self.make_form(schema=schema)
        self.assertEqual(form.fields, ['foo', 'bar'])

    def test_vue_tagname(self):
        form = self.make_form()
        self.assertEqual(form.vue_tagname, 'wutta-form')

    def test_vue_component(self):
        form = self.make_form()
        self.assertEqual(form.vue_component, 'WuttaForm')

    def test_contains(self):
        form = self.make_form(fields=['foo', 'bar'])
        self.assertIn('foo', form)
        self.assertNotIn('baz', form)

    def test_iter(self):
        form = self.make_form(fields=['foo', 'bar'])

        fields = list(iter(form))
        self.assertEqual(fields, ['foo', 'bar'])

        fields = []
        for field in form:
            fields.append(field)
        self.assertEqual(fields, ['foo', 'bar'])

    def test_set_fields(self):
        form = self.make_form(fields=['foo', 'bar'])
        self.assertEqual(form.fields, ['foo', 'bar'])
        form.set_fields(['baz'])
        self.assertEqual(form.fields, ['baz'])

    def test_append(self):
        form = self.make_form(fields=['one', 'two'])
        self.assertEqual(form.fields, ['one', 'two'])
        form.append('one', 'two', 'three')
        self.assertEqual(form.fields, ['one', 'two', 'three'])

    def test_remove(self):
        form = self.make_form(fields=['one', 'two', 'three', 'four'])
        self.assertEqual(form.fields, ['one', 'two', 'three', 'four'])
        form.remove('two', 'three')
        self.assertEqual(form.fields, ['one', 'four'])

    def test_set_node(self):
        form = self.make_form(fields=['foo', 'bar'])
        self.assertEqual(form.nodes, {})

        # complete node
        node = colander.SchemaNode(colander.Bool(), name='foo')
        form.set_node('foo', node)
        self.assertIs(form.nodes['foo'], node)

        # type only
        typ = colander.Bool()
        form.set_node('foo', typ)
        node = form.nodes['foo']
        self.assertIsInstance(node, colander.SchemaNode)
        self.assertIsInstance(node.typ, colander.Bool)
        self.assertEqual(node.name, 'foo')

        # schema is updated if already present
        schema = form.get_schema()
        self.assertIsNotNone(schema)
        typ = colander.Date()
        form.set_node('foo', typ)
        node = form.nodes['foo']
        self.assertIsInstance(node, colander.SchemaNode)
        self.assertIsInstance(node.typ, colander.Date)
        self.assertEqual(node.name, 'foo')

    def test_set_widget(self):
        form = self.make_form(fields=['foo', 'bar'])
        self.assertEqual(form.widgets, {})

        # basic
        widget = widgets.SelectWidget()
        form.set_widget('foo', widget)
        self.assertIs(form.widgets['foo'], widget)

        # schema is updated if already present
        schema = form.get_schema()
        self.assertIsNotNone(schema)
        self.assertIs(schema['foo'].widget, widget)
        new_widget = widgets.TextInputWidget()
        form.set_widget('foo', new_widget)
        self.assertIs(form.widgets['foo'], new_widget)
        self.assertIs(schema['foo'].widget, new_widget)

        # can also just specify widget pseudo-type (invalid)
        self.assertNotIn('bar', form.widgets)
        self.assertRaises(ValueError, form.set_widget, 'bar', 'ldjfadjfadj')

        # can also just specify widget pseudo-type (valid)
        self.assertNotIn('bar', form.widgets)
        form.set_widget('bar', 'notes')
        self.assertIsInstance(form.widgets['bar'], widgets.NotesWidget)

    def test_make_widget(self):
        form = self.make_form(fields=['foo', 'bar'])

        # notes
        widget = form.make_widget('notes')
        self.assertIsInstance(widget, widgets.NotesWidget)

        # invalid
        widget = form.make_widget('fdajvdafjjf')
        self.assertIsNone(widget)

    def test_set_default_widgets(self):
        model = self.app.model

        # no defaults for "plain" schema
        form = self.make_form(fields=['foo', 'bar'])
        self.assertEqual(form.widgets, {})

        # no defaults for "plain" mapped class
        form = self.make_form(model_class=model.Setting)
        self.assertEqual(form.widgets, {})

        class MyWidget(widgets.Widget):
            pass

        # widget set for datetime mapped field
        form = self.make_form(model_class=model.Upgrade)
        self.assertIn('created', form.widgets)
        self.assertIsNot(form.widgets['created'], MyWidget)
        self.assertNotIsInstance(form.widgets['created'], MyWidget)

        # widget *not* set for datetime, if override present
        form = self.make_form(model_class=model.Upgrade,
                              widgets={'created': MyWidget()})
        self.assertIn('created', form.widgets)
        self.assertIsInstance(form.widgets['created'], MyWidget)

        # mock up a table with all relevant column types
        class Whatever(model.Base):
            __tablename__ = 'whatever'
            id = sa.Column(sa.Integer(), primary_key=True)
            date = sa.Column(sa.Date())
            date_time = sa.Column(sa.DateTime())

        # widget set for all known types
        form = self.make_form(model_class=Whatever)
        self.assertIsInstance(form.widgets['date'], widgets.WuttaDateWidget)
        self.assertIsInstance(form.widgets['date_time'], widgets.WuttaDateTimeWidget)

    def test_set_grid(self):
        form = self.make_form(fields=['foo', 'bar'])
        self.assertNotIn('foo', form.widgets)
        self.assertNotIn('foogrid', form.grid_vue_context)

        grid = Grid(self.request, key='foogrid',
                    columns=['a', 'b'],
                    data=[{'a': 1, 'b': 2}, {'a': 3, 'b': 4}])

        form.set_grid('foo', grid)
        self.assertIn('foo', form.widgets)
        self.assertIsInstance(form.widgets['foo'], widgets.GridWidget)
        self.assertIn('foogrid', form.grid_vue_context)

    def test_set_validator(self):
        form = self.make_form(fields=['foo', 'bar'])
        self.assertEqual(form.validators, {})

        def validate1(node, value):
            pass

        # basic
        form.set_validator('foo', validate1)
        self.assertIs(form.validators['foo'], validate1)

        def validate2(node, value):
            pass

        # schema is updated if already present
        schema = form.get_schema()
        self.assertIsNotNone(schema)
        self.assertIs(schema['foo'].validator, validate1)
        form.set_validator('foo', validate2)
        self.assertIs(form.validators['foo'], validate2)
        self.assertIs(schema['foo'].validator, validate2)

    def test_set_default(self):
        form = self.make_form(fields=['foo', 'bar'])
        self.assertEqual(form.defaults, {})

        # basic
        form.set_default('foo', 42)
        self.assertEqual(form.defaults['foo'], 42)

    def test_get_schema(self):
        model = self.app.model
        form = self.make_form()
        self.assertIsNone(form.schema)

        # provided schema is returned
        schema = self.make_schema()
        form = self.make_form(schema=schema)
        self.assertIs(form.schema, schema)
        self.assertIs(form.get_schema(), schema)

        # schema is auto-generated if fields provided
        form = self.make_form(fields=['foo', 'bar'])
        schema = form.get_schema()
        self.assertEqual(len(schema.children), 2)
        self.assertEqual(schema['foo'].name, 'foo')

        # but auto-generating without fields is not supported
        form = self.make_form()
        self.assertIsNone(form.schema)
        self.assertRaises(NotImplementedError, form.get_schema)

        # schema is auto-generated if model_class provided
        form = self.make_form(model_class=model.Setting)
        schema = form.get_schema()
        self.assertEqual(len(schema.children), 2)
        self.assertIn('name', schema)
        self.assertIn('value', schema)

        # but node overrides are honored when auto-generating
        form = self.make_form(model_class=model.Setting)
        value_node = colander.SchemaNode(colander.Bool(), name='value')
        form.set_node('value', value_node)
        schema = form.get_schema()
        self.assertIs(schema['value'], value_node)

        # schema is auto-generated if model_instance provided
        form = self.make_form(model_instance=model.Setting(name='uhoh'))
        self.assertEqual(form.fields, ['name', 'value'])
        self.assertIsNone(form.schema)
        # nb. force method to get new fields
        del form.fields
        schema = form.get_schema()
        self.assertEqual(len(schema.children), 2)
        self.assertIn('name', schema)
        self.assertIn('value', schema)

        # ColanderAlchemy schema still has *all* requested fields
        form = self.make_form(model_instance=model.Setting(name='uhoh'),
                              fields=['name', 'value', 'foo', 'bar'])
        self.assertEqual(form.fields, ['name', 'value', 'foo', 'bar'])
        self.assertIsNone(form.schema)
        schema = form.get_schema()
        self.assertEqual(len(schema.children), 4)
        self.assertIn('name', schema)
        self.assertIn('value', schema)
        self.assertIn('foo', schema)
        self.assertIn('bar', schema)

        # schema nodes are required by default
        form = self.make_form(fields=['foo', 'bar'])
        schema = form.get_schema()
        self.assertIs(schema['foo'].missing, colander.required)
        self.assertIs(schema['bar'].missing, colander.required)

        # but fields can be marked *not* required
        form = self.make_form(fields=['foo', 'bar'])
        form.set_required('bar', False)
        schema = form.get_schema()
        self.assertIs(schema['foo'].missing, colander.required)
        self.assertIs(schema['bar'].missing, colander.null)

        # validator overrides are honored
        def validate(node, value): pass
        form = self.make_form(model_class=model.Setting)
        form.set_validator('name', validate)
        schema = form.get_schema()
        self.assertIs(schema['name'].validator, validate)

        # validator can be set for whole form
        form = self.make_form(model_class=model.Setting)
        schema = form.get_schema()
        self.assertIsNone(schema.validator)
        form = self.make_form(model_class=model.Setting)
        form.set_validator(None, validate)
        schema = form.get_schema()
        self.assertIs(schema.validator, validate)

        # default value overrides are honored
        form = self.make_form(model_class=model.Setting)
        form.set_default('name', 'foo')
        schema = form.get_schema()
        self.assertEqual(schema['name'].default, 'foo')

    def test_get_deform(self):
        model = self.app.model
        schema = self.make_schema()

        # basic
        form = self.make_form(schema=schema)
        self.assertFalse(hasattr(form, 'deform_form'))
        dform = form.get_deform()
        self.assertIsInstance(dform, deform.Form)
        self.assertIs(form.deform_form, dform)

        # with model instance as dict
        myobj = {'foo': 'one', 'bar': 'two'}
        form = self.make_form(schema=schema, model_instance=myobj)
        dform = form.get_deform()
        self.assertEqual(dform.cstruct, myobj)

        # with sqlalchemy model instance
        myobj = model.Setting(name='foo', value='bar')
        form = self.make_form(model_instance=myobj)
        dform = form.get_deform()
        self.assertEqual(dform.cstruct, {'name': 'foo', 'value': 'bar'})

        # sqlalchemy instance with null value
        myobj = model.Setting(name='foo', value=None)
        form = self.make_form(model_instance=myobj)
        dform = form.get_deform()
        self.assertEqual(dform.cstruct, {'name': 'foo', 'value': colander.null})

    def test_get_cancel_url(self):

        # is referrer by default
        form = self.make_form()
        self.request.get_referrer = MagicMock(return_value='/cancel-default')
        self.assertEqual(form.get_cancel_url(), '/cancel-default')
        del self.request.get_referrer

        # or can be static URL
        form = self.make_form(cancel_url='/cancel-static')
        self.assertEqual(form.get_cancel_url(), '/cancel-static')

        # or can be fallback URL (nb. 'NOPE' indicates no referrer)
        form = self.make_form(cancel_url_fallback='/cancel-fallback')
        self.request.get_referrer = MagicMock(return_value='NOPE')
        self.assertEqual(form.get_cancel_url(), '/cancel-fallback')
        del self.request.get_referrer

        # or can be referrer fallback, i.e. home page
        form = self.make_form()
        def get_referrer(default=None):
            if default == 'NOPE':
                return 'NOPE'
            return '/home-page'
        self.request.get_referrer = get_referrer
        self.assertEqual(form.get_cancel_url(), '/home-page')
        del self.request.get_referrer

    def test_get_label(self):
        form = self.make_form(fields=['foo', 'bar'])
        self.assertEqual(form.get_label('foo'), "Foo")
        form.set_label('foo', "Baz")
        self.assertEqual(form.get_label('foo'), "Baz")

    def test_set_label(self):
        form = self.make_form(fields=['foo', 'bar'])
        self.assertEqual(form.get_label('foo'), "Foo")
        form.set_label('foo', "Baz")
        self.assertEqual(form.get_label('foo'), "Baz")

        # schema should be updated when setting label
        schema = self.make_schema()
        form = self.make_form(schema=schema)
        form.set_label('foo', "Woohoo")
        self.assertEqual(form.get_label('foo'), "Woohoo")
        self.assertEqual(schema['foo'].title, "Woohoo")

    def test_readonly_fields(self):
        form = self.make_form(fields=['foo', 'bar'])
        self.assertEqual(form.readonly_fields, set())
        self.assertFalse(form.is_readonly('foo'))

        form.set_readonly('foo')
        self.assertEqual(form.readonly_fields, {'foo'})
        self.assertTrue(form.is_readonly('foo'))
        self.assertFalse(form.is_readonly('bar'))

        form.set_readonly('bar')
        self.assertEqual(form.readonly_fields, {'foo', 'bar'})
        self.assertTrue(form.is_readonly('foo'))
        self.assertTrue(form.is_readonly('bar'))

        form.set_readonly('foo', False)
        self.assertEqual(form.readonly_fields, {'bar'})
        self.assertFalse(form.is_readonly('foo'))
        self.assertTrue(form.is_readonly('bar'))

    def test_required_fields(self):
        form = self.make_form(fields=['foo', 'bar'])
        self.assertEqual(form.required_fields, {})
        self.assertIsNone(form.is_required('foo'))

        form.set_required('foo')
        self.assertEqual(form.required_fields, {'foo': True})
        self.assertTrue(form.is_required('foo'))
        self.assertIsNone(form.is_required('bar'))

        form.set_required('bar')
        self.assertEqual(form.required_fields, {'foo': True, 'bar': True})
        self.assertTrue(form.is_required('foo'))
        self.assertTrue(form.is_required('bar'))

        form.set_required('foo', False)
        self.assertEqual(form.required_fields, {'foo': False, 'bar': True})
        self.assertFalse(form.is_required('foo'))
        self.assertTrue(form.is_required('bar'))

    def test_render_vue_tag(self):
        schema = self.make_schema()
        form = self.make_form(schema=schema)
        html = form.render_vue_tag()
        self.assertEqual(html, '<wutta-form></wutta-form>')

    def test_render_vue_template(self):
        self.pyramid_config.include('pyramid_mako')
        self.pyramid_config.add_subscriber('wuttaweb.subscribers.before_render',
                                           'pyramid.events.BeforeRender')

        # form button is disabled on @submit by default
        schema = self.make_schema()
        form = self.make_form(schema=schema, cancel_url='/')
        html = form.render_vue_template()
        self.assertIn('<script type="text/x-template" id="wutta-form-template">', html)
        self.assertIn('@submit', html)

        # but not if form is configured otherwise
        form = self.make_form(schema=schema, auto_disable_submit=False, cancel_url='/')
        html = form.render_vue_template()
        self.assertIn('<script type="text/x-template" id="wutta-form-template">', html)
        self.assertNotIn('@submit', html)

    def test_add_grid_vue_context(self):
        form = self.make_form()

        # grid must have key
        grid = Grid(self.request)
        self.assertRaises(ValueError, form.add_grid_vue_context, grid)

        # otherwise it works
        grid = Grid(self.request, key='foo')
        self.assertEqual(len(form.grid_vue_context), 0)
        form.add_grid_vue_context(grid)
        self.assertEqual(len(form.grid_vue_context), 1)
        self.assertIn('foo', form.grid_vue_context)
        self.assertEqual(form.grid_vue_context['foo'], {
            'data': [],
            'row_classes': {},
        })

        # calling again with same key will replace data
        records = [{'foo': 1}, {'foo': 2}]
        grid = Grid(self.request, key='foo', columns=['foo'], data=records)
        form.add_grid_vue_context(grid)
        self.assertEqual(len(form.grid_vue_context), 1)
        self.assertIn('foo', form.grid_vue_context)
        self.assertEqual(form.grid_vue_context['foo'], {
            'data': records,
            'row_classes': {},
        })

    def test_render_vue_finalize(self):
        form = self.make_form()
        html = form.render_vue_finalize()
        self.assertIn('<script>', html)
        self.assertIn("Vue.component('wutta-form', WuttaForm)", html)

    def test_render_vue_field(self):
        self.pyramid_config.include('pyramid_deform')
        schema = self.make_schema()
        form = self.make_form(schema=schema)
        dform = form.get_deform()

        # typical
        html = form.render_vue_field('foo')
        self.assertIn('<b-field :horizontal="true" label="Foo">', html)
        self.assertIn('<b-input name="foo"', html)
        # nb. no error message
        self.assertNotIn('message', html)

        # readonly
        html = form.render_vue_field('foo', readonly=True)
        self.assertIn('<b-field :horizontal="true" label="Foo">', html)
        self.assertNotIn('<b-input name="foo"', html)
        # nb. no error message
        self.assertNotIn('message', html)

        # with error message
        with patch.object(form, 'get_field_errors', return_value=['something is wrong']):
            html = form.render_vue_field('foo')
            self.assertIn('something is wrong', html)

        # add another field, but not to deform, so it should still
        # display but with no widget
        form.fields.append('zanzibar')
        html = form.render_vue_field('zanzibar')
        self.assertIn('<b-field :horizontal="true" label="Zanzibar">', html)
        self.assertNotIn('<b-input', html)
        # nb. no error message
        self.assertNotIn('message', html)

        # try that once more but with a model record instance
        with patch.object(form, 'model_instance', new={'zanzibar': 'omgwtfbbq'}):
            html = form.render_vue_field('zanzibar')
        self.assertIn('<b-field', html)
        self.assertIn('label="Zanzibar"', html)
        self.assertNotIn('<b-input', html)
        self.assertIn('>omgwtfbbq<', html)
        # nb. no error message
        self.assertNotIn('message', html)

    def test_get_vue_field_value(self):
        schema = self.make_schema()
        form = self.make_form(schema=schema)

        # TODO: yikes what a hack (?)
        dform = form.get_deform()
        dform.set_appstruct({'foo': 'one', 'bar': 'two'})

        # null for missing field
        value = form.get_vue_field_value('doesnotexist')
        self.assertIsNone(value)

        # normal value is returned
        value = form.get_vue_field_value('foo')
        self.assertEqual(value, 'one')

        # but not if we remove field from deform
        # TODO: what is the use case here again?
        dform.children.remove(dform['foo'])
        value = form.get_vue_field_value('foo')
        self.assertIsNone(value)

    def test_get_vue_model_data(self):
        schema = self.make_schema()
        form = self.make_form(schema=schema)

        # 2 fields by default (foo, bar)
        data = form.get_vue_model_data()
        self.assertEqual(len(data), 2)

        # still just 2 fields even if we request more
        form.set_fields(['foo', 'bar', 'baz'])
        data = form.get_vue_model_data()
        self.assertEqual(len(data), 2)

        # confirm bool values make it thru as-is
        schema.add(colander.SchemaNode(colander.Bool(), name='baz'))
        form = self.make_form(schema=schema, model_instance={
            'foo': 'one',
            'bar': 'two',
            'baz': True,
        })
        data = form.get_vue_model_data()
        self.assertEqual(list(data.values()), ['one', 'two', True])

    def test_has_global_errors(self):

        def fail(node, value):
            node.raise_invalid("things are bad!")

        schema = self.make_schema()
        schema.validator = fail
        form = self.make_form(schema=schema)
        self.assertFalse(form.has_global_errors())
        self.request.method = 'POST'
        self.request.POST = {'foo': 'one', 'bar': 'two'}
        self.assertFalse(form.validate())
        self.assertTrue(form.has_global_errors())

    def test_get_global_errors(self):

        def fail(node, value):
            node.raise_invalid("things are bad!")

        schema = self.make_schema()
        schema.validator = fail
        form = self.make_form(schema=schema)
        self.assertEqual(form.get_global_errors(), [])
        self.request.method = 'POST'
        self.request.POST = {'foo': 'one', 'bar': 'two'}
        self.assertFalse(form.validate())
        self.assertTrue(form.get_global_errors(), ["things are bad!"])

    def test_get_field_errors(self):
        schema = self.make_schema()

        # simple 'Required' validation failure
        form = self.make_form(schema=schema)
        self.request.method = 'POST'
        self.request.POST = {'foo': 'one'}
        self.assertFalse(form.validate())
        errors = form.get_field_errors('bar')
        self.assertEqual(errors, ['Required'])

        # no errors
        form = self.make_form(schema=schema)
        self.request.POST = {'foo': 'one', 'bar': 'two'}
        self.assertTrue(form.validate())
        errors = form.get_field_errors('bar')
        self.assertEqual(errors, [])

    def test_validate(self):
        schema = self.make_schema()
        form = self.make_form(schema=schema)
        self.assertFalse(hasattr(form, 'validated'))

        # will not validate unless request is POST
        self.request.POST = {'foo': 'blarg', 'bar': 'baz'}
        self.request.method = 'GET'
        self.assertFalse(form.validate())
        self.request.method = 'POST'
        data = form.validate()
        self.assertEqual(data, {'foo': 'blarg', 'bar': 'baz'})

        # validating a second time updates form.validated
        self.request.POST = {'foo': 'BLARG', 'bar': 'BAZ'}
        data = form.validate()
        self.assertEqual(data, {'foo': 'BLARG', 'bar': 'BAZ'})
        self.assertIs(form.validated, data)

        # bad data does not validate
        self.request.POST = {'foo': 42, 'bar': None}
        self.assertFalse(form.validate())
        dform = form.get_deform()
        self.assertEqual(len(dform.error.children), 2)
        self.assertEqual(dform['foo'].errormsg, "Pstruct is not a string")

        # when a form has readonly fields, validating it will *remove*
        # those fields from deform/schema as well as final data dict
        schema = self.make_schema()
        form = self.make_form(schema=schema)
        form.set_readonly('foo')
        self.request.POST = {'foo': 'one', 'bar': 'two'}
        data = form.validate()
        self.assertEqual(data, {'bar': 'two'})
        dform = form.get_deform()
        self.assertNotIn('foo', schema)
        self.assertNotIn('foo', dform)
        self.assertIn('bar', schema)
        self.assertIn('bar', dform)



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
:term:`Email Settings <email setting>` for WuttaWeb
"""

from wuttjamaican.email import EmailSetting


class feedback(EmailSetting):
    """
    Sent when user submits feedback via the web app.
    """
    default_subject = "User Feedback"

    def sample_data(self):
        """ """
        model = self.app.model
        person = model.Person(full_name="Barney Rubble")
        user = model.User(username='barney', person=person)
        return {
            'user': user,
            'user_name': str(person),
            'user_url': '#',
            'referrer': 'http://example.com/',
            'client_ip': '127.0.0.1',
            'message': "This app is cool but needs a new feature.\n\nAllow me to describe...",
        }



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for app settings
"""

import json
import os
import sys
import subprocess
from collections import OrderedDict

from wuttjamaican.db.model import Setting
from wuttaweb.views import MasterView
from wuttaweb.util import get_libver, get_liburl


class AppInfoView(MasterView):
    """
    Master view for the core app info, to show/edit config etc.

    Default route prefix is ``appinfo``.

    Notable URLs provided by this class:

    * ``/appinfo/``
    * ``/appinfo/configure``

    See also :class:`SettingView`.
    """
    model_name = 'AppInfo'
    model_title_plural = "App Info"
    route_prefix = 'appinfo'
    filterable = False
    sort_on_backend = False
    sort_defaults = 'name'
    paginated = False
    creatable = False
    viewable = False
    editable = False
    deletable = False
    configurable = True

    grid_columns = [
        'name',
        'version',
        'editable_project_location',
    ]

    # TODO: for tailbone backward compat with get_liburl() etc.
    weblib_config_prefix = None

    def get_grid_data(self, columns=None, session=None):
        """ """

        # nb. init with empty data, only load it upon user request
        if not self.request.GET.get('partial'):
            return []

        # TODO: pretty sure this is not cross-platform.  probably some
        # sort of pip methods belong on the app handler?  or it should
        # have a pip handler for all that?
        pip = os.path.join(sys.prefix, 'bin', 'pip')
        output = subprocess.check_output([pip, 'list', '--format=json'], text=True)
        data = json.loads(output.strip())

        # must avoid null values for sort to work right
        for pkg in data:
            pkg.setdefault('editable_project_location', '')

        return data

    def configure_grid(self, g):
        """ """
        super().configure_grid(g)

        g.sort_multiple = False

        # name
        g.set_searchable('name')

        # editable_project_location
        g.set_searchable('editable_project_location')

    def get_weblibs(self):
        """ """
        return OrderedDict([
            ('vue', "(Vue2) Vue"),
            ('vue_resource', "(Vue2) vue-resource"),
            ('buefy', "(Vue2) Buefy"),
            ('buefy.css', "(Vue2) Buefy CSS"),
            ('fontawesome', "(Vue2) FontAwesome"),
            ('bb_vue', "(Vue3) vue"),
            ('bb_oruga', "(Vue3) @oruga-ui/oruga-next"),
            ('bb_oruga_bulma', "(Vue3) @oruga-ui/theme-bulma (JS)"),
            ('bb_oruga_bulma_css', "(Vue3) @oruga-ui/theme-bulma (CSS)"),
            ('bb_fontawesome_svg_core', "(Vue3) @fortawesome/fontawesome-svg-core"),
            ('bb_free_solid_svg_icons', "(Vue3) @fortawesome/free-solid-svg-icons"),
            ('bb_vue_fontawesome', "(Vue3) @fortawesome/vue-fontawesome"),
        ])

    def configure_get_simple_settings(self):
        """ """
        simple_settings = [

            # basics
            {'name': f'{self.config.appname}.app_title'},
            {'name': f'{self.config.appname}.node_type'},
            {'name': f'{self.config.appname}.node_title'},
            {'name': f'{self.config.appname}.production',
             'type': bool},
            {'name': f'{self.config.appname}.web.menus.handler.spec'},
            # nb. this is deprecated; we define so it is auto-deleted
            # when we replace with newer setting
            {'name': f'{self.config.appname}.web.menus.handler_spec'},

            # user/auth
            {'name': 'wuttaweb.home_redirect_to_login',
             'type': bool, 'default': False},

            # email
            {'name': f'{self.config.appname}.mail.send_emails',
             'type': bool, 'default': False},
            {'name': f'{self.config.appname}.email.default.sender'},
            {'name': f'{self.config.appname}.email.default.subject'},
            {'name': f'{self.config.appname}.email.default.to'},
            {'name': f'{self.config.appname}.email.feedback.subject'},
            {'name': f'{self.config.appname}.email.feedback.to'},

        ]

        def getval(key):
            return self.config.get(f'wuttaweb.{key}')

        weblibs = self.get_weblibs()
        for key, title in weblibs.items():

            simple_settings.append({
                'name': f'wuttaweb.libver.{key}',
                'default': getval(f'libver.{key}'),
            })
            simple_settings.append({
                'name': f'wuttaweb.liburl.{key}',
                'default': getval(f'liburl.{key}'),
            })

        return simple_settings

    def configure_get_context(self, **kwargs):
        """ """
        context = super().configure_get_context(**kwargs)

        # add registered menu handlers
        web = self.app.get_web_handler()
        handlers = web.get_menu_handler_specs()
        handlers = [{'spec': spec} for spec in handlers]
        context['menu_handlers'] = handlers

        # add `weblibs` to context, based on config values
        weblibs = self.get_weblibs()
        for key in weblibs:
            title = weblibs[key]
            weblibs[key] = {
                'key': key,
                'title': title,

                # nb. these values are exactly as configured, and are
                # used for editing the settings
                'configured_version': get_libver(self.request, key,
                                                 prefix=self.weblib_config_prefix,
                                                 configured_only=True),
                'configured_url': get_liburl(self.request, key,
                                             prefix=self.weblib_config_prefix,
                                             configured_only=True),

                # nb. these are for display only
                'default_version': get_libver(self.request, key,
                                              prefix=self.weblib_config_prefix,
                                              default_only=True),
                'live_url': get_liburl(self.request, key,
                                       prefix=self.weblib_config_prefix),
            }
        context['weblibs'] = list(weblibs.values())

        return context


class SettingView(MasterView):
    """
    Master view for the "raw" settings table.

    Default route prefix is ``settings``.

    Notable URLs provided by this class:

    * ``/settings/``

    See also :class:`AppInfoView`.
    """
    model_class = Setting
    model_title = "Raw Setting"
    deletable_bulk = True
    filter_defaults = {
        'name': {'active': True},
    }
    sort_defaults = 'name'

    # TODO: master should handle this (per model key)
    def configure_grid(self, g):
        """ """
        super().configure_grid(g)

        # name
        g.set_link('name')

    def configure_form(self, f):
        """ """
        super().configure_form(f)

        # name
        f.set_validator('name', self.unique_name)

        # value
        # TODO: master should handle this (per column nullable)
        f.set_required('value', False)

    def unique_name(self, node, value):
        """ """
        model = self.app.model
        session = self.Session()

        query = session.query(model.Setting)\
                       .filter(model.Setting.name == value)

        if self.editing:
            name = self.request.matchdict['name']
            query = query.filter(model.Setting.name != name)

        if query.count():
            node.raise_invalid("Setting name must be unique")


def defaults(config, **kwargs):
    base = globals()

    AppInfoView = kwargs.get('AppInfoView', base['AppInfoView'])
    AppInfoView.defaults(config)

    SettingView = kwargs.get('SettingView', base['SettingView'])
    SettingView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-

from unittest import TestCase

from pyramid import testing
from beaker.session import Session as BeakerSession

from wuttjamaican.testing import ConfigTestCase

from wuttaweb import progress as mod


class TestGetBasicSession(TestCase):

    def setUp(self):
        self.request = testing.DummyRequest()

    def test_basic(self):
        session = mod.get_basic_session(self.request)
        self.assertIsInstance(session, BeakerSession)
        self.assertFalse(session.use_cookies)


class TestGetProgressSession(TestCase):

    def setUp(self):
        self.request = testing.DummyRequest()

    def test_basic(self):
        self.request.session.id = 'mockid'
        session = mod.get_progress_session(self.request, 'foo')
        self.assertIsInstance(session, BeakerSession)
        self.assertEqual(session.id, 'mockid.progress.foo')


class TestSessionProgress(ConfigTestCase):

    def setUp(self):
        self.setup_config()
        self.request = testing.DummyRequest(wutta_config=self.config)
        self.request.session.id = 'mockid'

    def test_error_url(self):
        factory = mod.SessionProgress(self.request, 'foo', success_url='/blart')
        self.assertEqual(factory.error_url, '/blart')

    def test_basic(self):

        # sanity / coverage check
        factory = mod.SessionProgress(self.request, 'foo')
        prog = factory("doing things", 2)
        prog.update(1)
        prog.update(2)
        prog.handle_success()

    def test_error(self):

        # sanity / coverage check
        factory = mod.SessionProgress(self.request, 'foo')
        prog = factory("doing things", 2)
        prog.update(1)
        try:
            raise RuntimeError('omg')
        except Exception as error:
            prog.handle_error(error)



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for roles
"""

from wuttjamaican.db.model import Role, Permission
from wuttaweb.views import MasterView
from wuttaweb.db import Session
from wuttaweb.forms import widgets
from wuttaweb.forms.schema import UserRefs, Permissions, RoleRef


class RoleView(MasterView):
    """
    Master view for roles.

    Default route prefix is ``roles``.

    Notable URLs provided by this class:

    * ``/roles/``
    * ``/roles/new``
    * ``/roles/XXX``
    * ``/roles/XXX/edit``
    * ``/roles/XXX/delete``
    """
    model_class = Role

    grid_columns = [
        'name',
        'notes',
    ]

    filter_defaults = {
        'name': {'active': True},
    }
    sort_defaults = 'name'

    # TODO: master should handle this, possibly via configure_form()
    def get_query(self, session=None):
        """ """
        model = self.app.model
        query = super().get_query(session=session)
        return query.order_by(model.Role.name)

    def configure_grid(self, g):
        """ """
        super().configure_grid(g)

        # name
        g.set_link('name')

        # notes
        g.set_renderer('notes', self.grid_render_notes)

    def is_editable(self, role):
        """ """
        session = self.app.get_session(role)
        auth = self.app.get_auth_handler()

        # only "root" can edit admin role
        if role is auth.get_role_administrator(session):
            return self.request.is_root

        # other built-in roles require special perm
        if role in (auth.get_role_authenticated(session),
                    auth.get_role_anonymous(session)):
            return self.has_perm('edit_builtin')

        return True

    def is_deletable(self, role):
        """ """
        session = self.app.get_session(role)
        auth = self.app.get_auth_handler()

        # prevent delete for built-in roles
        if role is auth.get_role_authenticated(session):
            return False
        if role is auth.get_role_anonymous(session):
            return False
        if role is auth.get_role_administrator(session):
            return False

        return True

    def configure_form(self, f):
        """ """
        super().configure_form(f)
        role = f.model_instance

        # never show these
        f.remove('permission_refs',
                 'user_refs')

        # name
        f.set_validator('name', self.unique_name)

        # notes
        f.set_widget('notes', widgets.NotesWidget())

        # users
        if not (self.creating or self.editing):
            f.append('users')
            f.set_readonly('users')
            f.set_node('users', UserRefs(self.request))
            f.set_default('users', [u.uuid for u in role.users])

        # permissions
        f.append('permissions')
        self.wutta_permissions = self.get_available_permissions()
        f.set_node('permissions', Permissions(self.request, permissions=self.wutta_permissions))
        if not self.creating:
            f.set_default('permissions', list(role.permissions))

    def unique_name(self, node, value):
        """ """
        model = self.app.model
        session = Session()

        query = session.query(model.Role)\
                       .filter(model.Role.name == value)

        if self.editing:
            uuid = self.request.matchdict['uuid']
            query = query.filter(model.Role.uuid != uuid)

        if query.count():
            node.raise_invalid("Name must be unique")

    def get_available_permissions(self):
        """
        Returns all "available" permissions.  This is used when
        viewing or editing a role; the result is passed into the
        :class:`~wuttaweb.forms.schema.Permissions` field schema.

        The app itself must be made aware of each permission, in order
        for them to found by this method.  This is done via
        :func:`~wuttaweb.auth.add_permission_group()` and
        :func:`~wuttaweb.auth.add_permission()`.

        When in "view" (readonly) mode, this method will return the
        full set of known permissions.

        However in "edit" mode, it will prune the set to remove any
        permissions which the current user does not also have.  The
        idea here is to allow "many" users to manage roles, but ensure
        they cannot "break out" of their own role by assigning extra
        permissions to it.

        The permissions returned will also be grouped, and each single
        permission is also represented as a simple dict, e.g.::

           {
               'books': {
                   'key': 'books',
                   'label': "Books",
                   'perms': {
                       'books.list': {
                           'key': 'books.list',
                           'label': "Browse / search Books",
                       },
                       'books.view': {
                           'key': 'books.view',
                           'label': "View Book",
                       },
                   },
               },
               'widgets': {
                   'key': 'widgets',
                   'label': "Widgets",
                   'perms': {
                       'widgets.list': {
                           'key': 'widgets.list',
                           'label': "Browse / search Widgets",
                       },
                       'widgets.view': {
                           'key': 'widgets.view',
                           'label': "View Widget",
                       },
                   },
               },
           }
        """

        # get all known permissions from settings cache
        permissions = self.request.registry.settings.get('wutta_permissions', {})

        # when viewing, we allow all permissions to be exposed for all users
        if self.viewing:
            return permissions

        # admin user gets to manage all permissions
        if self.request.is_admin:
            return permissions

        # non-admin user can only see permissions they're granted
        available = {}
        for gkey, group in permissions.items():
            for pkey, perm in group['perms'].items():
                if self.request.has_perm(pkey):
                    if gkey not in available:
                        available[gkey] = {
                            'key': gkey,
                            'label': group['label'],
                            'perms': {},
                        }
                    available[gkey]['perms'][pkey] = perm

        return available

    def objectify(self, form):
        """ """
        # normal logic first
        role = super().objectify(form)

        # update permissions for role
        self.update_permissions(role, form)

        return role

    def update_permissions(self, role, form):
        """ """
        if 'permissions' not in form.validated:
            return

        auth = self.app.get_auth_handler()
        available = self.wutta_permissions
        permissions = form.validated['permissions']

        for gkey, group in available.items():
            for pkey, perm in group['perms'].items():
                if pkey in permissions:
                    auth.grant_permission(role, pkey)
                else:
                    auth.revoke_permission(role, pkey)

    @classmethod
    def defaults(cls, config):
        """ """
        cls._defaults(config)
        cls._role_defaults(config)

    @classmethod
    def _role_defaults(cls, config):
        permission_prefix = cls.get_permission_prefix()
        model_title_plural = cls.get_model_title_plural()

        # perm to edit built-in roles
        config.add_wutta_permission(permission_prefix,
                                    f'{permission_prefix}.edit_builtin',
                                    f"Edit the Built-in {model_title_plural}")


class PermissionView(MasterView):
    """
    Master view for permissions.

    Default route prefix is ``permissions``.

    Notable URLs provided by this class:

    * ``/permissions/``
    * ``/permissions/XXX``
    * ``/permissions/XXX/delete``
    """
    model_class = Permission
    creatable = False
    editable = False

    grid_columns = [
        'role',
        'permission',
    ]

    sort_defaults = 'role'

    form_fields = [
        'role',
        'permission',
    ]

    def get_query(self, **kwargs):
        """ """
        query = super().get_query(**kwargs)
        model = self.app.model

        # always join on Role
        query = query.join(model.Role)

        return query

    def configure_grid(self, g):
        """ """
        super().configure_grid(g)
        model = self.app.model

        # role
        g.set_sorter('role', model.Role.name)
        g.set_filter('role', model.Role.name, label="Role Name")
        g.set_link('role')

        # permission
        g.set_link('permission')

    def configure_form(self, f):
        """ """
        super().configure_form(f)

        # role
        f.set_node('role', RoleRef(self.request))


def defaults(config, **kwargs):
    base = globals()

    RoleView = kwargs.get('RoleView', base['RoleView'])
    RoleView.defaults(config)

    PermissionView = kwargs.get('PermissionView', base['PermissionView'])
    PermissionView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-

import datetime
import decimal
from unittest import TestCase
from unittest.mock import patch

import colander
from pyramid import testing

from sqlalchemy import orm

from wuttjamaican.conf import WuttaConfig
from wuttaweb.forms import schema as mod
from wuttaweb.forms import widgets
from wuttaweb.testing import DataTestCase, WebTestCase


class TestWuttaDateTime(TestCase):

    def test_deserialize(self):
        typ = mod.WuttaDateTime()
        node = colander.SchemaNode(typ)

        result = typ.deserialize(node, colander.null)
        self.assertIs(result, colander.null)

        result = typ.deserialize(node, '2024-12-11T10:33 PM')
        self.assertIsInstance(result, datetime.datetime)
        self.assertEqual(result, datetime.datetime(2024, 12, 11, 22, 33))
        self.assertIsNone(result.tzinfo)

        result = typ.deserialize(node, '2024-12-11T22:33:00')
        self.assertIsInstance(result, datetime.datetime)
        self.assertEqual(result, datetime.datetime(2024, 12, 11, 22, 33))
        self.assertIsNone(result.tzinfo)

        self.assertRaises(colander.Invalid, typ.deserialize, node, 'bogus')


class TestObjectNode(DataTestCase):

    def setUp(self):
        self.setup_db()
        self.request = testing.DummyRequest(wutta_config=self.config)

    def test_dictify(self):
        model = self.app.model
        person = model.Person(full_name="Betty Boop")

        # unsupported type is converted to string
        node = mod.ObjectNode(colander.String())
        value = node.dictify(person)
        self.assertEqual(value, "Betty Boop")

        # but supported type can dictify
        node = mod.ObjectNode(mod.PersonRef(self.request))
        value = node.dictify(person)
        self.assertIs(value, person)

    def test_objectify(self):
        model = self.app.model
        person = model.Person(full_name="Betty Boop")

        # unsupported type raises error
        node = mod.ObjectNode(colander.String())
        self.assertRaises(NotImplementedError, node.objectify, person)

        # but supported type can objectify
        node = mod.ObjectNode(mod.PersonRef(self.request))
        value = node.objectify(person)
        self.assertIs(value, person)


class TestWuttaEnum(WebTestCase):

    def test_widget_maker(self):
        enum = self.app.enum
        typ = mod.WuttaEnum(self.request, enum.UpgradeStatus)
        widget = typ.widget_maker()
        self.assertIsInstance(widget, widgets.SelectWidget)


MOCK_STATUS_ONE = 1
MOCK_STATUS_TWO = 2
MOCK_STATUS = {
    MOCK_STATUS_ONE: 'one',
    MOCK_STATUS_TWO: 'two',
}

class TestWuttaDictEnum(WebTestCase):

    def test_widget_maker(self):
        typ = mod.WuttaDictEnum(self.request, MOCK_STATUS)
        widget = typ.widget_maker()
        self.assertIsInstance(widget, widgets.SelectWidget)
        self.assertEqual(widget.values, [
            (1, 'one'),
            (2, 'two'),
        ])


class TestWuttaMoney(WebTestCase):

    def test_widget_maker(self):
        enum = self.app.enum

        # default scale
        typ = mod.WuttaMoney(self.request)
        widget = typ.widget_maker()
        self.assertIsInstance(widget, widgets.WuttaMoneyInputWidget)
        self.assertEqual(widget.scale, 2)

        # custom scale
        typ = mod.WuttaMoney(self.request, scale=4)
        widget = typ.widget_maker()
        self.assertIsInstance(widget, widgets.WuttaMoneyInputWidget)
        self.assertEqual(widget.scale, 4)


class TestWuttaQuantity(WebTestCase):

    def test_serialize(self):
        node = colander.SchemaNode(mod.WuttaQuantity(self.request))
        typ = node.typ

        # null
        result = typ.serialize(node, colander.null)
        self.assertIs(result, colander.null)
        result = typ.serialize(node, None)
        self.assertIs(result, colander.null)

        # quantity
        result = typ.serialize(node, 42)
        self.assertEqual(result, '42')
        result = typ.serialize(node, 42.00)
        self.assertEqual(result, '42')
        result = typ.serialize(node, decimal.Decimal('42.00'))
        self.assertEqual(result, '42')
        result = typ.serialize(node, 42.13)
        self.assertEqual(result, '42.13')


class TestObjectRef(DataTestCase):

    def setUp(self):
        self.setup_db()
        self.request = testing.DummyRequest(wutta_config=self.config)

    def test_empty_option(self):

        # null by default
        typ = mod.ObjectRef(self.request)
        self.assertIsNone(typ.empty_option)

        # passing true yields default empty option
        typ = mod.ObjectRef(self.request, empty_option=True)
        self.assertEqual(typ.empty_option, ('', "(none)"))

        # can set explicitly
        typ = mod.ObjectRef(self.request, empty_option=('foo', 'bar'))
        self.assertEqual(typ.empty_option, ('foo', 'bar'))

        # can set just a label
        typ = mod.ObjectRef(self.request, empty_option="(empty)")
        self.assertEqual(typ.empty_option, ('', "(empty)"))

    def test_model_class(self):
        typ = mod.ObjectRef(self.request)
        self.assertRaises(NotImplementedError, getattr, typ, 'model_class')

    def test_serialize(self):
        model = self.app.model
        node = colander.SchemaNode(colander.String())

        # null
        typ = mod.ObjectRef(self.request)
        value = typ.serialize(node, colander.null)
        self.assertIs(value, colander.null)

        # model instance
        person = model.Person(full_name="Betty Boop")
        self.session.add(person)
        self.session.commit()
        self.assertIsNotNone(person.uuid)
        typ = mod.ObjectRef(self.request)
        value = typ.serialize(node, person)
        self.assertEqual(value, person.uuid.hex)

        # null w/ empty option
        typ = mod.ObjectRef(self.request, empty_option=('bad', 'BAD'))
        value = typ.serialize(node, colander.null)
        self.assertEqual(value, 'bad')

    def test_deserialize(self):
        model = self.app.model
        node = colander.SchemaNode(colander.String())

        # null
        typ = mod.ObjectRef(self.request)
        value = typ.deserialize(node, colander.null)
        self.assertIs(value, colander.null)

        # model instance
        person = model.Person(full_name="Betty Boop")
        self.session.add(person)
        self.session.commit()
        self.assertIsNotNone(person.uuid)
        with patch.object(mod.ObjectRef, 'model_class', new=model.Person):
            with patch.object(mod, 'Session', return_value=self.session):
                typ = mod.ObjectRef(self.request)
                value = typ.deserialize(node, person.uuid)
                self.assertIs(value, person)

    def test_dictify(self):
        model = self.app.model
        node = colander.SchemaNode(colander.String())

        # model instance
        person = model.Person(full_name="Betty Boop")
        self.session.add(person)
        self.session.commit()
        self.assertIsNotNone(person.uuid)
        typ = mod.ObjectRef(self.request)
        value = typ.dictify(person)
        self.assertIs(value, person)

    def test_objectify(self):
        model = self.app.model
        node = colander.SchemaNode(colander.String())

        # null
        typ = mod.ObjectRef(self.request)
        value = typ.objectify(None)
        self.assertIsNone(value)

        with patch.object(mod, 'Session', return_value=self.session):

            # model instance
            person = model.Person(full_name="Betty Boop")
            self.session.add(person)
            self.session.commit()
            self.assertIsNotNone(person.uuid)
            with patch.object(mod.ObjectRef, 'model_class', new=model.Person):

                # can specify as uuid
                typ = mod.ObjectRef(self.request)
                value = typ.objectify(person.uuid)
                self.assertIs(value, person)

                # or can specify object proper
                typ = mod.ObjectRef(self.request)
                value = typ.objectify(person)
                self.assertIs(value, person)

            # error if not found
            with patch.object(mod.ObjectRef, 'model_class', new=model.Person):
                typ = mod.ObjectRef(self.request)
                self.assertRaises(ValueError, typ.objectify, 'WRONG-UUID')

    def test_get_query(self):
        model = self.app.model
        with patch.object(mod.ObjectRef, 'model_class', new=model.Person):
            with patch.object(mod, 'Session', return_value=self.session):
                typ = mod.ObjectRef(self.request)
                query = typ.get_query()
                self.assertIsInstance(query, orm.Query)

    def test_sort_query(self):
        model = self.app.model
        with patch.object(mod.ObjectRef, 'model_class', new=model.Person):
            with patch.object(mod, 'Session', return_value=self.session):
                typ = mod.ObjectRef(self.request)
                query = typ.get_query()
                sorted_query = typ.sort_query(query)
                self.assertIs(sorted_query, query)

    def test_widget_maker(self):
        model = self.app.model
        person = model.Person(full_name="Betty Boop")
        self.session.add(person)
        self.session.commit()

        # basic
        with patch.object(mod.ObjectRef, 'model_class', new=model.Person):
            with patch.object(mod, 'Session', return_value=self.session):
                typ = mod.ObjectRef(self.request)
                widget = typ.widget_maker()
                self.assertEqual(len(widget.values), 1)
                self.assertEqual(widget.values[0][1], "Betty Boop")

        # empty option
        with patch.object(mod.ObjectRef, 'model_class', new=model.Person):
            with patch.object(mod, 'Session', return_value=self.session):
                typ = mod.ObjectRef(self.request, empty_option=True)
                widget = typ.widget_maker()
                self.assertEqual(len(widget.values), 2)
                self.assertEqual(widget.values[0][1], "(none)")
                self.assertEqual(widget.values[1][1], "Betty Boop")


class TestPersonRef(WebTestCase):

    def test_sort_query(self):
        with patch.object(mod, 'Session', return_value=self.session):
            typ = mod.PersonRef(self.request)
            query = typ.get_query()
            self.assertIsInstance(query, orm.Query)
            sorted_query = typ.sort_query(query)
            self.assertIsInstance(sorted_query, orm.Query)
            self.assertIsNot(sorted_query, query)

    def test_get_object_url(self):
        self.pyramid_config.add_route('people.view', '/people/{uuid}')
        model = self.app.model
        with patch.object(mod, 'Session', return_value=self.session):
            typ = mod.PersonRef(self.request)

            person = model.Person(full_name="Barney Rubble")
            self.session.add(person)
            self.session.commit()

            url = typ.get_object_url(person)
            self.assertIsNotNone(url)
            self.assertIn(f'/people/{person.uuid}', url)


class TestRoleRef(WebTestCase):

    def test_sort_query(self):
        with patch.object(mod, 'Session', return_value=self.session):
            typ = mod.RoleRef(self.request)
            query = typ.get_query()
            self.assertIsInstance(query, orm.Query)
            sorted_query = typ.sort_query(query)
            self.assertIsInstance(sorted_query, orm.Query)
            self.assertIsNot(sorted_query, query)

    def test_get_object_url(self):
        self.pyramid_config.add_route('roles.view', '/roles/{uuid}')
        model = self.app.model
        with patch.object(mod, 'Session', return_value=self.session):
            typ = mod.RoleRef(self.request)

            role = model.Role(name='Manager')
            self.session.add(role)
            self.session.commit()

            url = typ.get_object_url(role)
            self.assertIsNotNone(url)
            self.assertIn(f'/roles/{role.uuid}', url)


class TestUserRef(WebTestCase):

    def test_sort_query(self):
        with patch.object(mod, 'Session', return_value=self.session):
            typ = mod.UserRef(self.request)
            query = typ.get_query()
            self.assertIsInstance(query, orm.Query)
            sorted_query = typ.sort_query(query)
            self.assertIsInstance(sorted_query, orm.Query)
            self.assertIsNot(sorted_query, query)

    def test_get_object_url(self):
        self.pyramid_config.add_route('users.view', '/users/{uuid}')
        model = self.app.model
        with patch.object(mod, 'Session', return_value=self.session):
            typ = mod.UserRef(self.request)

            user = model.User(username='barney')
            self.session.add(user)
            self.session.commit()

            url = typ.get_object_url(user)
            self.assertIsNotNone(url)
            self.assertIn(f'/users/{user.uuid}', url)


class TestUserRefs(DataTestCase):

    def setUp(self):
        self.setup_db()
        self.request = testing.DummyRequest(wutta_config=self.config)

    def test_widget_maker(self):
        model = self.app.model
        with patch.object(mod, 'Session', return_value=self.session):
            typ = mod.UserRefs(self.request)
            widget = typ.widget_maker()
            self.assertIsInstance(widget, widgets.UserRefsWidget)


class TestRoleRefs(DataTestCase):

    def setUp(self):
        self.setup_db()
        self.request = testing.DummyRequest(wutta_config=self.config)

    def test_widget_maker(self):
        model = self.app.model
        auth = self.app.get_auth_handler()
        admin = auth.get_role_administrator(self.session)
        authed = auth.get_role_authenticated(self.session)
        anon = auth.get_role_anonymous(self.session)
        blokes = model.Role(name="Blokes")
        self.session.add(blokes)
        self.session.commit()

        with patch.object(mod, 'Session', return_value=self.session):

            # with root access, default values include: admin, blokes
            self.request.is_root = True
            typ = mod.RoleRefs(self.request)
            widget = typ.widget_maker()
            self.assertEqual(len(widget.values), 2)
            self.assertEqual(widget.values[0][1], "Administrator")
            self.assertEqual(widget.values[1][1], "Blokes")

            # without root, default values include: blokes
            self.request.is_root = False
            typ = mod.RoleRefs(self.request)
            widget = typ.widget_maker()
            self.assertEqual(len(widget.values), 1)
            self.assertEqual(widget.values[0][1], "Blokes")


class TestPermissions(DataTestCase):

    def setUp(self):
        self.setup_db()
        self.request = testing.DummyRequest(wutta_config=self.config)

    def test_widget_maker(self):

        # no supported permissions
        permissions = {}
        typ = mod.Permissions(self.request, permissions)
        widget = typ.widget_maker()
        self.assertEqual(len(widget.values), 0)

        # supported permissions are morphed to values
        permissions = {
            'widgets': {
                'label': "Widgets",
                'perms': {
                    'widgets.polish': {
                        'label': "Polish the widgets",
                    },
                },
            },
        }
        typ = mod.Permissions(self.request, permissions)
        widget = typ.widget_maker()
        self.assertEqual(len(widget.values), 1)
        self.assertEqual(widget.values[0], ('widgets.polish', "Polish the widgets"))


class TestFileDownload(DataTestCase):

    def setUp(self):
        self.setup_db()
        self.request = testing.DummyRequest(wutta_config=self.config)

    def test_widget_maker(self):

        # sanity / coverage check
        typ = mod.FileDownload(self.request, url='/foo')
        widget = typ.widget_maker()
        self.assertIsInstance(widget, widgets.FileDownloadWidget)
        self.assertEqual(widget.url, '/foo')


class TestEmailRecipients(TestCase):

    def test_serialize(self):
        typ = mod.EmailRecipients()
        node = colander.SchemaNode(typ)

        recips = [
            'alice@example.com',
            'bob@example.com',
        ]
        recips_str = ', '.join(recips)

        # values
        result = typ.serialize(node, recips_str)
        self.assertEqual(result, '\n'.join(recips))

        # null
        result = typ.serialize(node, colander.null)
        self.assertIs(result, colander.null)

    def test_deserialize(self):
        typ = mod.EmailRecipients()
        node = colander.SchemaNode(typ)

        recips = [
            'alice@example.com',
            'bob@example.com',
        ]
        recips_str = ', '.join(recips)

        # values
        result = typ.deserialize(node, recips_str)
        self.assertEqual(result, recips_str)

        # null
        result = typ.deserialize(node, colander.null)
        self.assertIs(result, colander.null)

    def test_widget_maker(self):
        typ = mod.EmailRecipients()
        widget = typ.widget_maker()
        self.assertIsInstance(widget, widgets.EmailRecipientsWidget)



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Auth Views
"""

import colander

from wuttaweb.views import View
from wuttaweb.db import Session
from wuttaweb.auth import login_user, logout_user
from wuttaweb.forms import widgets


class AuthView(View):
    """
    Auth views shared by all apps.
    """

    def login(self, session=None):
        """
        View for user login.

        This view shows the login form, and handles its submission.
        Upon successful login, user is redirected to home page.

        * route: ``login``
        * template: ``/auth/login.mako``
        """
        model = self.app.model
        session = session or Session()
        auth = self.app.get_auth_handler()

        # nb. redirect to /setup if no users exist
        user = session.query(model.User).first()
        if not user:
            return self.redirect(self.request.route_url('setup'))

        referrer = self.request.get_referrer()

        # redirect if already logged in
        if self.request.user:
            self.request.session.flash(f"{self.request.user} is already logged in", 'error')
            return self.redirect(referrer)

        form = self.make_form(schema=self.login_make_schema(),
                              align_buttons_right=True,
                              show_button_cancel=False,
                              show_button_reset=True,
                              button_label_submit="Login",
                              button_icon_submit='user')

        # validate basic form data (sanity check)
        data = form.validate()
        if data:

            # truly validate user credentials
            user = auth.authenticate_user(session, data['username'], data['password'])
            if user:

                # okay now they're truly logged in
                headers = login_user(self.request, user)
                return self.redirect(referrer, headers=headers)

            else:
                self.request.session.flash("Invalid user credentials", 'error')

        return {
            'index_title': self.app.get_title(),
            'form': form,
            # TODO
            # 'referrer': referrer,
        }

    def login_make_schema(self):
        schema = colander.Schema()

        # nb. we must explicitly declare the widgets in order to also
        # specify the ref attribute.  this is needed for autofocus and
        # keydown behavior for login form.

        schema.add(colander.SchemaNode(
            colander.String(),
            name='username',
            widget=widgets.TextInputWidget(attributes={
                'ref': 'username',
            })))

        schema.add(colander.SchemaNode(
            colander.String(),
            name='password',
            widget=widgets.PasswordWidget(attributes={
                'ref': 'password',
            })))

        return schema

    def logout(self):
        """
        View for user logout.

        This deletes/invalidates the current user session and then
        redirects to the login page.

        Note that a simple GET is sufficient; POST is not required.

        * route: ``logout``
        * template: n/a
        """
        # truly logout the user
        headers = logout_user(self.request)

        # TODO
        # # redirect to home page after logout, if so configured
        # if self.config.get_bool('wuttaweb.home_after_logout', default=False):
        #     return self.redirect(self.request.route_url('home'), headers=headers)

        # otherwise redirect to referrer, with 'login' page as fallback
        # TODO: should call request.get_referrer()
        # referrer = self.request.get_referrer(default=self.request.route_url('login'))
        referrer = self.request.route_url('login')
        return self.redirect(referrer, headers=headers)

    def change_password(self):
        """
        View allowing a user to change their own password.

        This view shows a change-password form, and handles its
        submission.  If successful, user is redirected to home page.

        If current user is not authenticated, no form is shown and
        user is redirected to home page.

        * route: ``change_password``
        * template: ``/auth/change_password.mako``
        """
        if not self.request.user:
            return self.redirect(self.request.route_url('home'))

        if self.request.user.prevent_edit:
            raise self.forbidden()

        form = self.make_form(schema=self.change_password_make_schema(),
                              show_button_cancel=False,
                              show_button_reset=True)

        data = form.validate()
        if data:
            auth = self.app.get_auth_handler()
            auth.set_user_password(self.request.user, data['new_password'])
            self.request.session.flash("Your password has been changed.")
            # TODO: should use request.get_referrer() instead
            referrer = self.request.route_url('home')
            return self.redirect(referrer)

        return {'index_title': str(self.request.user),
                'form': form}

    def change_password_make_schema(self):
        schema = colander.Schema()

        schema.add(colander.SchemaNode(
            colander.String(),
            name='current_password',
            widget=widgets.PasswordWidget(),
            validator=self.change_password_validate_current_password))

        schema.add(colander.SchemaNode(
            colander.String(),
            name='new_password',
            widget=widgets.CheckedPasswordWidget(),
            validator=self.change_password_validate_new_password))

        return schema

    def change_password_validate_current_password(self, node, value):
        auth = self.app.get_auth_handler()
        user = self.request.user
        if not auth.check_user_password(user, value):
            node.raise_invalid("Current password is incorrect.")

    def change_password_validate_new_password(self, node, value):
        auth = self.app.get_auth_handler()
        user = self.request.user
        if auth.check_user_password(user, value):
            node.raise_invalid("New password must be different from old password.")

    def become_root(self):
        """
        Elevate the current request to 'root' for full system access.

        This is only allowed if current (authenticated) user is a
        member of the Administrator role.  Also note that GET is not
        allowed for this view, only POST.

        See also :meth:`stop_root()`.
        """
        if self.request.method != 'POST':
            raise self.forbidden()

        if not self.request.is_admin:
            raise self.forbidden()

        self.request.session['is_root'] = True
        self.request.session.flash("You have been elevated to 'root' and now have full system access")

        url = self.request.get_referrer()
        return self.redirect(url)

    def stop_root(self):
        """
        Lower the current request from 'root' back to normal access.

        Also note that GET is not allowed for this view, only POST.

        See also :meth:`become_root()`.
        """
        if self.request.method != 'POST':
            raise self.forbidden()

        if not self.request.is_admin:
            raise self.forbidden()

        self.request.session['is_root'] = False
        self.request.session.flash("Your normal system access has been restored")

        url = self.request.get_referrer()
        return self.redirect(url)

    @classmethod
    def defaults(cls, config):
        cls._auth_defaults(config)

    @classmethod
    def _auth_defaults(cls, config):

        # login
        config.add_route('login', '/login')
        config.add_view(cls, attr='login',
                        route_name='login',
                        renderer='/auth/login.mako')

        # logout
        config.add_route('logout', '/logout')
        config.add_view(cls, attr='logout',
                        route_name='logout')

        # change password
        config.add_route('change_password', '/change-password')
        config.add_view(cls, attr='change_password',
                        route_name='change_password',
                        renderer='/auth/change_password.mako')

        # become root
        config.add_route('become_root', '/root/yes',
                         request_method='POST')
        config.add_view(cls, attr='become_root',
                        route_name='become_root')

        # stop root
        config.add_route('stop_root', '/root/no',
                         request_method='POST')
        config.add_view(cls, attr='stop_root',
                        route_name='stop_root')


def defaults(config, **kwargs):
    base = globals()

    AuthView = kwargs.get('AuthView', base['AuthView'])
    AuthView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-

import datetime
import decimal
import functools
from unittest import TestCase
from unittest.mock import MagicMock, patch

from sqlalchemy import orm
from pyramid import testing
from pyramid.response import Response
from pyramid.httpexceptions import HTTPNotFound

from wuttjamaican.conf import WuttaConfig
from wuttaweb.views import master as mod
from wuttaweb.views import View
from wuttaweb.progress import SessionProgress
from wuttaweb.subscribers import new_request_set_user
from wuttaweb.testing import WebTestCase


class TestMasterView(WebTestCase):

    def make_view(self):
        return mod.MasterView(self.request)

    def test_defaults(self):
        with patch.multiple(mod.MasterView, create=True,
                            model_name='Widget',
                            model_key='uuid',
                            deletable_bulk=True,
                            has_autocomplete=True,
                            downloadable=True,
                            executable=True,
                            configurable=True):
            mod.MasterView.defaults(self.pyramid_config)

    ##############################
    # class methods
    ##############################

    def test_get_model_class(self):
        
        # no model class by default
        self.assertIsNone(mod.MasterView.get_model_class())

        # subclass may specify
        MyModel = MagicMock()
        with patch.multiple(mod.MasterView, create=True,
                            model_class=MyModel):
            self.assertIs(mod.MasterView.get_model_class(), MyModel)

    def test_get_model_name(self):
        
        # error by default (since no model class)
        self.assertRaises(AttributeError, mod.MasterView.get_model_name)

        # subclass may specify model name
        with patch.multiple(mod.MasterView, create=True,
                            model_name='Widget'):
            self.assertEqual(mod.MasterView.get_model_name(), 'Widget')

        # or it may specify model class
        MyModel = MagicMock(__name__='Blaster')
        with patch.multiple(mod.MasterView, create=True,
                            model_class=MyModel):
            self.assertEqual(mod.MasterView.get_model_name(), 'Blaster')

    def test_get_model_name_normalized(self):
        
        # error by default (since no model class)
        self.assertRaises(AttributeError, mod.MasterView.get_model_name_normalized)

        # subclass may specify *normalized* model name
        with patch.multiple(mod.MasterView, create=True,
                            model_name_normalized='widget'):
            self.assertEqual(mod.MasterView.get_model_name_normalized(), 'widget')

        # or it may specify *standard* model name
        with patch.multiple(mod.MasterView, create=True,
                            model_name='Blaster'):
            self.assertEqual(mod.MasterView.get_model_name_normalized(), 'blaster')

        # or it may specify model class
        MyModel = MagicMock(__name__='Dinosaur')
        with patch.multiple(mod.MasterView, create=True,
                            model_class=MyModel):
            self.assertEqual(mod.MasterView.get_model_name_normalized(), 'dinosaur')

    def test_get_model_title(self):
        
        # error by default (since no model class)
        self.assertRaises(AttributeError, mod.MasterView.get_model_title)

        # subclass may specify  model title
        with patch.multiple(mod.MasterView, create=True,
                            model_title='Wutta Widget'):
            self.assertEqual(mod.MasterView.get_model_title(), "Wutta Widget")

        # or it may specify model name
        with patch.multiple(mod.MasterView, create=True,
                            model_name='Blaster'):
            self.assertEqual(mod.MasterView.get_model_title(), "Blaster")

        # or it may specify model class
        MyModel = MagicMock(__name__='Dinosaur')
        with patch.multiple(mod.MasterView, create=True,
                            model_class=MyModel):
            self.assertEqual(mod.MasterView.get_model_title(), "Dinosaur")

    def test_get_model_title_plural(self):
        
        # error by default (since no model class)
        self.assertRaises(AttributeError, mod.MasterView.get_model_title_plural)

        # subclass may specify *plural* model title
        with patch.multiple(mod.MasterView, create=True,
                            model_title_plural='People'):
            self.assertEqual(mod.MasterView.get_model_title_plural(), "People")

        # or it may specify *singular* model title
        with patch.multiple(mod.MasterView, create=True,
                            model_title='Wutta Widget'):
            self.assertEqual(mod.MasterView.get_model_title_plural(), "Wutta Widgets")

        # or it may specify model name
        with patch.multiple(mod.MasterView, create=True,
                            model_name='Blaster'):
            self.assertEqual(mod.MasterView.get_model_title_plural(), "Blasters")

        # or it may specify model class
        MyModel = MagicMock(__name__='Dinosaur')
        with patch.multiple(mod.MasterView, create=True,
                            model_class=MyModel):
            self.assertEqual(mod.MasterView.get_model_title_plural(), "Dinosaurs")

    def test_get_model_key(self):

        # error by default (since no model class)
        self.assertRaises(AttributeError, mod.MasterView.get_model_key)

        # subclass may specify model key
        with patch.multiple(mod.MasterView, create=True,
                            model_key='uuid'):
            self.assertEqual(mod.MasterView.get_model_key(), ('uuid',))

    def test_get_route_prefix(self):
        
        # error by default (since no model class)
        self.assertRaises(AttributeError, mod.MasterView.get_route_prefix)

        # subclass may specify route prefix
        with patch.multiple(mod.MasterView, create=True,
                            route_prefix='widgets'):
            self.assertEqual(mod.MasterView.get_route_prefix(), 'widgets')

        # subclass may specify *normalized* model name
        with patch.multiple(mod.MasterView, create=True,
                            model_name_normalized='blaster'):
            self.assertEqual(mod.MasterView.get_route_prefix(), 'blasters')

        # or it may specify *standard* model name
        with patch.multiple(mod.MasterView, create=True,
                            model_name = 'Dinosaur'):
            self.assertEqual(mod.MasterView.get_route_prefix(), 'dinosaurs')

        # or it may specify model class
        MyModel = MagicMock(__name__='Truck')
        with patch.multiple(mod.MasterView, create=True,
                            model_class=MyModel):
            self.assertEqual(mod.MasterView.get_route_prefix(), 'trucks')

    def test_get_permission_prefix(self):

        # error by default (since no model class)
        self.assertRaises(AttributeError, mod.MasterView.get_permission_prefix)

        # subclass may specify permission prefix
        with patch.object(mod.MasterView, 'permission_prefix', new='widgets', create=True):
            self.assertEqual(mod.MasterView.get_permission_prefix(), 'widgets')

        # subclass may specify route prefix
        with patch.object(mod.MasterView, 'route_prefix', new='widgets', create=True):
            self.assertEqual(mod.MasterView.get_permission_prefix(), 'widgets')

        # or it may specify model class
        Truck = MagicMock(__name__='Truck')
        with patch.object(mod.MasterView, 'model_class', new=Truck, create=True):
            self.assertEqual(mod.MasterView.get_permission_prefix(), 'trucks')

    def test_get_url_prefix(self):
        
        # error by default (since no model class)
        self.assertRaises(AttributeError, mod.MasterView.get_url_prefix)

        # subclass may specify url prefix
        with patch.multiple(mod.MasterView, create=True,
                            url_prefix='/widgets'):
            self.assertEqual(mod.MasterView.get_url_prefix(), '/widgets')

        # or it may specify route prefix
        with patch.multiple(mod.MasterView, create=True,
                            route_prefix='trucks'):
            self.assertEqual(mod.MasterView.get_url_prefix(), '/trucks')

        # or it may specify *normalized* model name
        with patch.multiple(mod.MasterView, create=True,
                            model_name_normalized='blaster'):
            self.assertEqual(mod.MasterView.get_url_prefix(), '/blasters')

        # or it may specify *standard* model name
        with patch.multiple(mod.MasterView, create=True,
                            model_name='Dinosaur'):
            self.assertEqual(mod.MasterView.get_url_prefix(), '/dinosaurs')

        # or it may specify model class
        MyModel = MagicMock(__name__='Machine')
        with patch.multiple(mod.MasterView, create=True,
                            model_class=MyModel):
            self.assertEqual(mod.MasterView.get_url_prefix(), '/machines')

    def test_get_instance_url_prefix(self):

        # error by default (since no model class)
        self.assertRaises(AttributeError, mod.MasterView.get_instance_url_prefix)

        # typical example with url_prefix and simple key
        with patch.multiple(mod.MasterView, create=True,
                            url_prefix='/widgets',
                            model_key='uuid'):
            self.assertEqual(mod.MasterView.get_instance_url_prefix(), '/widgets/{uuid}')

        # typical example with composite key
        with patch.multiple(mod.MasterView, create=True,
                            url_prefix='/widgets',
                            model_key=('foo', 'bar')):
            self.assertEqual(mod.MasterView.get_instance_url_prefix(), '/widgets/{foo}|{bar}')

    def test_get_template_prefix(self):
        
        # error by default (since no model class)
        self.assertRaises(AttributeError, mod.MasterView.get_template_prefix)

        # subclass may specify template prefix
        with patch.multiple(mod.MasterView, create=True,
                            template_prefix='/widgets'):
            self.assertEqual(mod.MasterView.get_template_prefix(), '/widgets')

        # or it may specify url prefix
        with patch.multiple(mod.MasterView, create=True,
                            url_prefix='/trees'):
            self.assertEqual(mod.MasterView.get_template_prefix(), '/trees')

        # or it may specify route prefix
        with patch.multiple(mod.MasterView, create=True,
                            route_prefix='trucks'):
            self.assertEqual(mod.MasterView.get_template_prefix(), '/trucks')

        # or it may specify *normalized* model name
        with patch.multiple(mod.MasterView, create=True,
                            model_name_normalized='blaster'):
            self.assertEqual(mod.MasterView.get_template_prefix(), '/blasters')

        # or it may specify *standard* model name
        with patch.multiple(mod.MasterView, create=True,
                            model_name='Dinosaur'):
            self.assertEqual(mod.MasterView.get_template_prefix(), '/dinosaurs')

        # or it may specify model class
        MyModel = MagicMock(__name__='Machine')
        with patch.multiple(mod.MasterView, create=True,
                            model_class=MyModel):
            self.assertEqual(mod.MasterView.get_template_prefix(), '/machines')

    def test_get_grid_key(self):

        # error by default (since no model class)
        self.assertRaises(AttributeError, mod.MasterView.get_grid_key)

        # subclass may specify grid key
        with patch.multiple(mod.MasterView, create=True,
                            grid_key='widgets'):
            self.assertEqual(mod.MasterView.get_grid_key(), 'widgets')

        # or it may specify route prefix
        with patch.multiple(mod.MasterView, create=True,
                            route_prefix='trucks'):
            self.assertEqual(mod.MasterView.get_grid_key(), 'trucks')

        # or it may specify *normalized* model name
        with patch.multiple(mod.MasterView, create=True,
                            model_name_normalized='blaster'):
            self.assertEqual(mod.MasterView.get_grid_key(), 'blasters')

        # or it may specify *standard* model name
        with patch.multiple(mod.MasterView, create=True,
                            model_name='Dinosaur'):
            self.assertEqual(mod.MasterView.get_grid_key(), 'dinosaurs')

        # or it may specify model class
        MyModel = MagicMock(__name__='Machine')
        with patch.multiple(mod.MasterView, create=True,
                            model_class=MyModel):
            self.assertEqual(mod.MasterView.get_grid_key(), 'machines')

    def test_get_config_title(self):

        # error by default (since no model class)
        self.assertRaises(AttributeError, mod.MasterView.get_config_title)

        # subclass may specify config title
        with patch.multiple(mod.MasterView, create=True,
                            config_title='Widgets'):
            self.assertEqual(mod.MasterView.get_config_title(), "Widgets")

        # subclass may specify *plural* model title
        with patch.multiple(mod.MasterView, create=True,
                            model_title_plural='People'):
            self.assertEqual(mod.MasterView.get_config_title(), "People")

        # or it may specify *singular* model title
        with patch.multiple(mod.MasterView, create=True,
                            model_title='Wutta Widget'):
            self.assertEqual(mod.MasterView.get_config_title(), "Wutta Widgets")

        # or it may specify model name
        with patch.multiple(mod.MasterView, create=True,
                            model_name='Blaster'):
            self.assertEqual(mod.MasterView.get_config_title(), "Blasters")

        # or it may specify model class
        MyModel = MagicMock(__name__='Dinosaur')
        with patch.multiple(mod.MasterView, create=True,
                            model_class=MyModel):
            self.assertEqual(mod.MasterView.get_config_title(), "Dinosaurs")

    def test_get_row_model_class(self):
        model = self.app.model

        # no default
        self.assertIsNone(mod.MasterView.get_row_model_class())

        # class may specify
        with patch.object(mod.MasterView, 'row_model_class', create=True, new=model.User):
            self.assertIs(mod.MasterView.get_row_model_class(), model.User)

    ##############################
    # support methods
    ##############################

    def test_get_class_hierarchy(self):
        class MyView(mod.MasterView):
            pass

        view = MyView(self.request)
        classes = view.get_class_hierarchy()
        self.assertEqual(classes, [View, mod.MasterView, MyView])

    def test_has_perm(self):
        model = self.app.model
        auth = self.app.get_auth_handler()

        with patch.multiple(mod.MasterView, create=True,
                            model_name='Setting'):
            view = self.make_view()

            # anonymous user
            self.assertFalse(view.has_perm('list'))
            self.assertFalse(self.request.has_perm('list'))

            # reset
            del self.request.user_permissions

            # make user with perms
            barney = model.User(username='barney')
            self.session.add(barney)
            blokes = model.Role(name="Blokes")
            self.session.add(blokes)
            barney.roles.append(blokes)
            auth.grant_permission(blokes, 'settings.list')
            self.session.commit()

            # this user has perms
            self.request.user = barney
            self.assertTrue(view.has_perm('list'))
            self.assertTrue(self.request.has_perm('settings.list'))

    def test_has_any_perm(self):
        model = self.app.model
        auth = self.app.get_auth_handler()

        with patch.multiple(mod.MasterView, create=True,
                            model_name='Setting'):
            view = self.make_view()

            # anonymous user
            self.assertFalse(view.has_any_perm('list', 'view'))
            self.assertFalse(self.request.has_any_perm('settings.list', 'settings.view'))

            # reset
            del self.request.user_permissions

            # make user with perms
            barney = model.User(username='barney')
            self.session.add(barney)
            blokes = model.Role(name="Blokes")
            self.session.add(blokes)
            barney.roles.append(blokes)
            auth.grant_permission(blokes, 'settings.view')
            self.session.commit()

            # this user has perms
            self.request.user = barney
            self.assertTrue(view.has_any_perm('list', 'view'))
            self.assertTrue(self.request.has_any_perm('settings.list', 'settings.view'))

    def test_make_button(self):
        view = self.make_view()

        # normal
        html = view.make_button('click me')
        self.assertIn('<b-button ', html)
        self.assertIn('click me', html)
        self.assertNotIn('is-primary', html)

        # primary as primary
        html = view.make_button('click me', primary=True)
        self.assertIn('<b-button ', html)
        self.assertIn('click me', html)
        self.assertIn('is-primary', html)

        # primary as variant
        html = view.make_button('click me', variant='is-primary')
        self.assertIn('<b-button ', html)
        self.assertIn('click me', html)
        self.assertIn('is-primary', html)

        # primary as type
        html = view.make_button('click me', type='is-primary')
        self.assertIn('<b-button ', html)
        self.assertIn('click me', html)
        self.assertIn('is-primary', html)

        # with url
        html = view.make_button('click me', url='http://example.com')
        self.assertIn('<b-button tag="a"', html)
        self.assertIn('click me', html)
        self.assertIn('href="http://example.com"', html)

    def test_make_progress(self):

        # basic
        view = self.make_view()
        self.request.session.id = 'mockid'
        progress = view.make_progress('foo')
        self.assertIsInstance(progress, SessionProgress)

    def test_render_progress(self):
        self.pyramid_config.add_route('progress', '/progress/{key}')

        # sanity / coverage check
        view = self.make_view()
        progress = MagicMock()
        response = view.render_progress(progress)

    def test_render_to_response(self):
        self.pyramid_config.include('wuttaweb.views.common')
        self.pyramid_config.include('wuttaweb.views.auth')
        self.pyramid_config.add_route('appinfo', '/appinfo/')

        def widgets(request): return {}
        self.pyramid_config.add_route('widgets', '/widgets/')
        self.pyramid_config.add_view(widgets, route_name='widgets')

        # basic sanity check using /master/index.mako
        # (nb. it skips /widgets/index.mako since that doesn't exist)
        with patch.multiple(mod.MasterView, create=True,
                            model_name='Widget',
                            creatable=False):
            view = mod.MasterView(self.request)
            response = view.render_to_response('index', {})
            self.assertIsInstance(response, Response)

        # basic sanity check using /appinfo/index.mako
        with patch.multiple(mod.MasterView, create=True,
                            model_name='AppInfo',
                            route_prefix='appinfo',
                            url_prefix='/appinfo',
                            creatable=False):
            view = mod.MasterView(self.request)
            response = view.render_to_response('index', {
                # nb. grid is required for this template
                'grid': MagicMock(),
            })
            self.assertIsInstance(response, Response)

        # bad template name causes error
        with patch.multiple(mod.MasterView, create=True,
                            model_name='Widget'):
            self.assertRaises(IOError, view.render_to_response, 'nonexistent', {})

    def test_get_index_title(self):
        with patch.multiple(mod.MasterView, create=True,
                            model_title_plural = "Wutta Widgets"):
            view = mod.MasterView(self.request)
            self.assertEqual(view.get_index_title(), "Wutta Widgets")

    def test_collect_labels(self):

        # no labels by default
        view = self.make_view()
        labels = view.collect_labels()
        self.assertEqual(labels, {})

        # labels come from all classes; subclass wins
        with patch.object(View, 'labels', new={'foo': "Foo", 'bar': "Bar"}, create=True):
            with patch.object(mod.MasterView, 'labels', new={'foo': "FOO FIGHTERS"}, create=True):
                view = self.make_view()
                labels = view.collect_labels()
                self.assertEqual(labels, {'foo': "FOO FIGHTERS", 'bar': "Bar"})

    def test_set_labels(self):
        model = self.app.model
        with patch.object(mod.MasterView, 'model_class', new=model.Setting, create=True):

            # no labels by default
            view = self.make_view()
            grid = view.make_model_grid(session=self.session)
            view.set_labels(grid)
            self.assertEqual(grid.labels, {})

            # labels come from all classes; subclass wins
            with patch.object(mod.MasterView, 'labels', new={'name': "SETTING NAME"}, create=True):
                view = self.make_view()
                view.set_labels(grid)
                self.assertEqual(grid.labels, {'name': "SETTING NAME"})

    def test_make_model_grid(self):
        self.pyramid_config.add_route('settings.delete_bulk', '/settings/delete-bulk')
        model = self.app.model

        # no model class
        with patch.multiple(mod.MasterView, create=True,
                            model_name='Widget',
                            model_key='uuid'):
            view = mod.MasterView(self.request)
            grid = view.make_model_grid()
            self.assertIsNone(grid.model_class)

        # explicit model class
        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting):
            grid = view.make_model_grid(session=self.session)
            self.assertIs(grid.model_class, model.Setting)

        # no row class by default
        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting):
            grid = view.make_model_grid(session=self.session)
            self.assertIsNone(grid.row_class)

        # can specify row class
        get_row_class = MagicMock()
        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting,
                            grid_row_class=get_row_class):
            grid = view.make_model_grid(session=self.session)
            self.assertIs(grid.row_class, get_row_class)

        # no actions by default
        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting):
            grid = view.make_model_grid(session=self.session)
            self.assertEqual(grid.actions, [])

        # now let's test some more actions logic
        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting,
                            viewable=True,
                            editable=True,
                            deletable=True):

            # should have 3 actions now, but for lack of perms
            grid = view.make_model_grid(session=self.session)
            self.assertEqual(len(grid.actions), 0)

            # but root user has perms, so gets 3 actions
            with patch.object(self.request, 'is_root', new=True):
                grid = view.make_model_grid(session=self.session)
            self.assertEqual(len(grid.actions), 3)

        # no tools by default
        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting):
            grid = view.make_model_grid(session=self.session)
            self.assertEqual(grid.tools, {})

        # delete-results tool added if master/perms allow
        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting,
                            deletable_bulk=True):
            with patch.object(self.request, 'is_root', new=True):
                grid = view.make_model_grid(session=self.session)
                self.assertIn('delete-results', grid.tools)

    def test_get_grid_data(self):
        model = self.app.model
        self.app.save_setting(self.session, 'foo', 'bar')
        self.session.commit()
        setting = self.session.query(model.Setting).one()
        view = self.make_view()

        # empty by default
        self.assertFalse(hasattr(mod.MasterView, 'model_class'))
        data = view.get_grid_data(session=self.session)
        self.assertEqual(data, [])

        # grid with model class will produce data query
        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting):
            view = mod.MasterView(self.request)
            query = view.get_grid_data(session=self.session)
            self.assertIsInstance(query, orm.Query)
            data = query.all()
            self.assertEqual(len(data), 1)
            self.assertIs(data[0], setting)

    def test_configure_grid(self):
        model = self.app.model

        # uuid field is pruned
        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting):
            view = mod.MasterView(self.request)
            grid = view.make_grid(model_class=model.Setting,
                                  columns=['uuid', 'name', 'value'])
            self.assertIn('uuid', grid.columns)
            view.configure_grid(grid)
            self.assertNotIn('uuid', grid.columns)

    def test_grid_render_bool(self):
        model = self.app.model
        view = self.make_view()
        user = model.User(username='barney', active=None)

        # null
        value = view.grid_render_bool(user, 'active', None)
        self.assertIsNone(value)

        # true
        user.active = True
        value = view.grid_render_bool(user, 'active', True)
        self.assertEqual(value, "Yes")

        # false
        user.active = False
        value = view.grid_render_bool(user, 'active', False)
        self.assertEqual(value, "No")

    def test_grid_render_currency(self):
        view = self.make_view()
        obj = {'amount': None}

        # null
        value = view.grid_render_currency(obj, 'amount', None)
        self.assertIsNone(value)

        # normal amount
        obj['amount'] = decimal.Decimal('100.42')
        value = view.grid_render_currency(obj, 'amount', '100.42')
        self.assertEqual(value, "$100.42")

        # negative amount
        obj['amount'] = decimal.Decimal('-100.42')
        value = view.grid_render_currency(obj, 'amount', '-100.42')
        self.assertEqual(value, "($100.42)")

    def test_grid_render_datetime(self):
        view = self.make_view()
        obj = {'dt': None}

        # null
        value = view.grid_render_datetime(obj, 'dt', None)
        self.assertIsNone(value)

        # normal
        obj['dt'] = datetime.datetime(2024, 8, 24, 11)
        value = view.grid_render_datetime(obj, 'dt', '2024-08-24T11:00:00')
        self.assertEqual(value, '2024-08-24 11:00:00 AM')

    def test_grid_render_enum(self):
        enum = self.app.enum
        view = self.make_view()
        obj = {'status': None}

        # null
        value = view.grid_render_enum(obj, 'status', None, enum=enum.UpgradeStatus)
        self.assertIsNone(value)

        # normal
        obj['status'] = enum.UpgradeStatus.SUCCESS
        value = view.grid_render_enum(obj, 'status', 'SUCCESS', enum=enum.UpgradeStatus)
        self.assertEqual(value, 'SUCCESS')

    def test_grid_render_notes(self):
        model = self.app.model
        view = self.make_view()

        # null
        text = None
        role = model.Role(name="Foo", notes=text)
        value = view.grid_render_notes(role, 'notes', text)
        self.assertIsNone(value)

        # short string
        text = "hello world"
        role = model.Role(name="Foo", notes=text)
        value = view.grid_render_notes(role, 'notes', text)
        self.assertEqual(value, text)

        # long string
        text = "hello world " * 20
        role = model.Role(name="Foo", notes=text)
        value = view.grid_render_notes(role, 'notes', text)
        self.assertIn('<span ', value)

    def test_get_instance(self):
        model = self.app.model
        self.app.save_setting(self.session, 'foo', 'bar')
        self.session.commit()
        self.assertEqual(self.session.query(model.Setting).count(), 1)

        # default not implemented
        view = mod.MasterView(self.request)
        self.assertRaises(NotImplementedError, view.get_instance)

        # fetch from DB if model class is known
        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting):
            view = mod.MasterView(self.request)

            # existing setting is returned
            self.request.matchdict = {'name': 'foo'}
            setting = view.get_instance(session=self.session)
            self.assertIsInstance(setting, model.Setting)
            self.assertEqual(setting.name, 'foo')
            self.assertEqual(setting.value, 'bar')

            # missing setting not found
            self.request.matchdict = {'name': 'blarg'}
            self.assertRaises(HTTPNotFound, view.get_instance, session=self.session)

    def test_get_action_route_kwargs(self):
        model = self.app.model
        with patch.object(mod.MasterView, 'model_class', new=model.Setting, create=True):
            view = self.make_view()

            # dict object
            setting = {'name': 'foo', 'value': 'bar'}
            kw = view.get_action_route_kwargs(setting)
            self.assertEqual(kw, {'name': 'foo'})

            # mapped object
            setting = model.Setting(name='foo', value='bar')
            kw = view.get_action_route_kwargs(setting)
            self.assertEqual(kw, {'name': 'foo'})

            # non-standard object
            class MySetting:
                def __init__(self, **kw):
                    self.__dict__.update(kw)
            setting = MySetting(name='foo', value='bar')
            kw = view.get_action_route_kwargs(setting)
            self.assertEqual(kw, {'name': 'foo'})

    def test_get_action_url_for_dict(self):
        model = self.app.model
        setting = {'name': 'foo', 'value': 'bar'}
        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting):
            mod.MasterView.defaults(self.pyramid_config)
            view = self.make_view()
            url = view.get_action_url_view(setting, 0)
            self.assertEqual(url, self.request.route_url('settings.view', name='foo'))

    def test_get_action_url_for_orm_object(self):
        model = self.app.model
        setting = model.Setting(name='foo', value='bar')
        self.session.add(setting)
        self.session.commit()
        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting):
            mod.MasterView.defaults(self.pyramid_config)
            view = self.make_view()
            url = view.get_action_url_view(setting, 0)
            self.assertEqual(url, self.request.route_url('settings.view', name='foo'))

    def test_get_action_url_for_adhoc_object(self):
        model = self.app.model
        class MockSetting:
            def __init__(self, **kw):
                self.__dict__.update(kw)
        setting = MockSetting(name='foo', value='bar')
        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting):
            mod.MasterView.defaults(self.pyramid_config)
            view = self.make_view()
            url = view.get_action_url_view(setting, 0)
            self.assertEqual(url, self.request.route_url('settings.view', name='foo'))

    def test_get_action_url_view(self):
        model = self.app.model
        setting = model.Setting(name='foo', value='bar')
        self.session.add(setting)
        self.session.commit()

        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting):
            mod.MasterView.defaults(self.pyramid_config)
            view = self.make_view()
            url = view.get_action_url_view(setting, 0)
            self.assertEqual(url, self.request.route_url('settings.view', name='foo'))

    def test_get_action_url_edit(self):
        model = self.app.model
        setting = model.Setting(name='foo', value='bar')
        self.session.add(setting)
        self.session.commit()
        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting):
            mod.MasterView.defaults(self.pyramid_config)
            view = self.make_view()

            # typical
            url = view.get_action_url_edit(setting, 0)
            self.assertEqual(url, self.request.route_url('settings.edit', name='foo'))

            # but null if instance not editable
            with patch.object(view, 'is_editable', return_value=False):
                url = view.get_action_url_edit(setting, 0)
                self.assertIsNone(url)

    def test_get_action_url_delete(self):
        model = self.app.model
        setting = model.Setting(name='foo', value='bar')
        self.session.add(setting)
        self.session.commit()
        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting):
            mod.MasterView.defaults(self.pyramid_config)
            view = self.make_view()

            # typical
            url = view.get_action_url_delete(setting, 0)
            self.assertEqual(url, self.request.route_url('settings.delete', name='foo'))

            # but null if instance not deletable
            with patch.object(view, 'is_deletable', return_value=False):
                url = view.get_action_url_delete(setting, 0)
                self.assertIsNone(url)

    def test_make_model_form(self):
        model = self.app.model

        # no model class
        with patch.multiple(mod.MasterView, create=True,
                            model_name='Widget',
                            model_key='uuid'):
            view = mod.MasterView(self.request)
            form = view.make_model_form()
            self.assertIsNone(form.model_class)

        # explicit model class
        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting):
            form = view.make_model_form()
            self.assertIs(form.model_class, model.Setting)

    def test_configure_form(self):
        model = self.app.model

        # uuid field is pruned
        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting):
            view = mod.MasterView(self.request)
            form = view.make_form(model_class=model.Setting,
                                  fields=['uuid', 'name', 'value'])
            self.assertIn('uuid', form.fields)
            view.configure_form(form)
            self.assertNotIn('uuid', form.fields)

    def test_objectify(self):
        model = self.app.model
        self.app.save_setting(self.session, 'foo', 'bar')
        self.session.commit()
        self.assertEqual(self.session.query(model.Setting).count(), 1)

        # no model class
        with patch.multiple(mod.MasterView, create=True,
                            model_name='Widget',
                            model_key='uuid'):
            view = mod.MasterView(self.request)
            form = view.make_model_form(fields=['name', 'description'])
            form.validated = {'name': 'first'}
            obj = view.objectify(form)
            self.assertIs(obj, form.validated)

        # explicit model class (editing)
        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting,
                            editing=True):
            form = view.make_model_form()
            form.validated = {'name': 'foo', 'value': 'blarg'}
            form.model_instance = self.session.query(model.Setting).one()
            obj = view.objectify(form)
            self.assertIsInstance(obj, model.Setting)
            self.assertEqual(obj.name, 'foo')
            self.assertEqual(obj.value, 'blarg')

        # explicit model class (creating)
        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting,
                            creating=True):
            form = view.make_model_form()
            form.validated = {'name': 'another', 'value': 'whatever'}
            obj = view.objectify(form)
            self.assertIsInstance(obj, model.Setting)
            self.assertEqual(obj.name, 'another')
            self.assertEqual(obj.value, 'whatever')

    def test_persist(self):
        model = self.app.model
        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting):
            view = mod.MasterView(self.request)

            # new instance is persisted
            setting = model.Setting(name='foo', value='bar')
            self.assertEqual(self.session.query(model.Setting).count(), 0)
            view.persist(setting, session=self.session)
            self.session.commit()
            setting = self.session.query(model.Setting).one()
            self.assertEqual(setting.name, 'foo')
            self.assertEqual(setting.value, 'bar')

    ##############################
    # view methods
    ##############################

    def test_index(self):
        self.pyramid_config.include('wuttaweb.views.common')
        self.pyramid_config.include('wuttaweb.views.auth')
        self.pyramid_config.add_route('settings.create', '/settings/new')
        self.pyramid_config.add_route('settings.view', '/settings/{name}')
        self.pyramid_config.add_route('settings.edit', '/settings/{name}/edit')
        self.pyramid_config.add_route('settings.delete', '/settings/{name}/delete')
        
        # sanity/coverage check using /settings/
        with patch.multiple(mod.MasterView, create=True,
                            model_name='Setting',
                            model_key='name',
                            get_index_url=MagicMock(return_value='/settings/'),
                            grid_columns=['name', 'value']):
            view = mod.MasterView(self.request)
            response = view.index()

            # then again with data, to include view action url
            data = [{'name': 'foo', 'value': 'bar'}]
            with patch.object(view, 'get_grid_data', return_value=data):
                response = view.index()
                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.content_type, 'text/html')

                # then once more as 'partial' - aka. data only
                self.request.GET = {'partial': '1'}
                response = view.index()
                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.content_type, 'application/json')

                # redirects when view is reset
                self.request.GET = {'reset-view': '1', 'hash': 'foo'}
                with patch.object(self.request, 'current_route_url'):
                    response = view.index()
                self.assertEqual(response.status_code, 302)

    def test_create(self):
        self.pyramid_config.include('wuttaweb.views.common')
        self.pyramid_config.include('wuttaweb.views.auth')
        self.pyramid_config.add_route('settings.view', '/settings/{name}')
        model = self.app.model

        # sanity/coverage check using /settings/new
        with patch.multiple(mod.MasterView, create=True,
                            model_name='Setting',
                            model_key='name',
                            get_index_url=MagicMock(return_value='/settings/'),
                            form_fields=['name', 'value']):
            view = mod.MasterView(self.request)

            # no setting yet
            self.assertIsNone(self.app.get_setting(self.session, 'foo.bar'))

            # get the form page
            response = view.create()
            self.assertIsInstance(response, Response)
            self.assertEqual(response.status_code, 200)
            # self.assertIn('frazzle', response.text)
            # nb. no error
            self.assertNotIn('Required', response.text)

            def persist(setting):
                self.app.save_setting(self.session, setting['name'], setting['value'])
                self.session.commit()

            # post request to save setting
            self.request.method = 'POST'
            self.request.POST = {
                'name': 'foo.bar',
                'value': 'fraggle',
            }
            with patch.object(view, 'persist', new=persist):
                response = view.create()
            # nb. should get redirect back to view page
            self.assertEqual(response.status_code, 302)
            # setting should now be in DB
            self.assertEqual(self.app.get_setting(self.session, 'foo.bar'), 'fraggle')

            # try another post with invalid data (value is required)
            self.request.method = 'POST'
            self.request.POST = {}
            with patch.object(view, 'persist', new=persist):
                response = view.create()
            # nb. should get a form with errors
            self.assertEqual(response.status_code, 200)
            self.assertIn('Required', response.text)
            # setting did not change in DB
            self.assertEqual(self.app.get_setting(self.session, 'foo.bar'), 'fraggle')

    def test_view(self):
        self.pyramid_config.include('wuttaweb.views.common')
        self.pyramid_config.include('wuttaweb.views.auth')
        self.pyramid_config.add_route('settings.create', '/settings/new')
        self.pyramid_config.add_route('settings.edit', '/settings/{name}/edit')
        self.pyramid_config.add_route('settings.delete', '/settings/{name}/delete')

        # sanity/coverage check using /settings/XXX
        setting = {'name': 'foo.bar', 'value': 'baz'}
        self.request.matchdict = {'name': 'foo.bar'}
        with patch.multiple(mod.MasterView, create=True,
                            model_name='Setting',
                            model_key='name',
                            get_index_url=MagicMock(return_value='/settings/'),
                            grid_columns=['name', 'value'],
                            form_fields=['name', 'value']):
            view = mod.MasterView(self.request)
            with patch.object(view, 'get_instance', return_value=setting):
                response = view.view()

    def test_view_with_rows(self):
        self.pyramid_config.include('wuttaweb.views.common')
        self.pyramid_config.include('wuttaweb.views.auth')
        self.pyramid_config.add_route('people', '/people/')
        model = self.app.model
        person = model.Person(full_name="Whitney Houston")
        self.session.add(person)
        user = model.User(username='whitney', person=person)
        self.session.add(user)
        self.session.commit()

        get_row_grid_data = MagicMock()
        with patch.multiple(mod.MasterView, create=True,
                            Session=MagicMock(return_value=self.session),
                            model_class=model.Person,
                            route_prefix='people',
                            has_rows=True,
                            row_model_class=model.User,
                            get_row_grid_data=get_row_grid_data):
            with patch.object(self.request, 'matchdict', new={'uuid': person.uuid}):
                view = self.make_view()

                # just for coverage
                get_row_grid_data.return_value = []
                response = view.view()
                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.content_type, 'text/html')

                # now with data...
                get_row_grid_data.return_value = [user]
                response = view.view()
                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.content_type, 'text/html')

                # then once more as 'partial' - aka. data only
                with patch.dict(self.request.GET, {'partial': 1}):
                    response = view.view()
                    self.assertEqual(response.status_code, 200)
                    self.assertEqual(response.content_type, 'application/json')

                # redirects when view is reset
                with patch.dict(self.request.GET, {'reset-view': '1', 'hash': 'foo'}):
                    # nb. mock current route
                    with patch.object(self.request, 'current_route_url'):
                        response = view.view()
                        self.assertEqual(response.status_code, 302)

    def test_edit(self):
        self.pyramid_config.include('wuttaweb.views.common')
        self.pyramid_config.include('wuttaweb.views.auth')
        self.pyramid_config.add_route('settings.create', '/settings/new')
        self.pyramid_config.add_route('settings.view', '/settings/{name}')
        self.pyramid_config.add_route('settings.delete', '/settings/{name}/delete')
        model = self.app.model
        self.app.save_setting(self.session, 'foo.bar', 'frazzle')
        self.session.commit()

        def get_instance():
            setting = self.session.get(model.Setting, 'foo.bar')
            return {
                'name': setting.name,
                'value': setting.value,
            }

        # sanity/coverage check using /settings/XXX/edit
        self.request.matchdict = {'name': 'foo.bar'}
        with patch.multiple(mod.MasterView, create=True,
                            model_name='Setting',
                            model_key='name',
                            get_index_url=MagicMock(return_value='/settings/'),
                            form_fields=['name', 'value']):
            view = mod.MasterView(self.request)
            with patch.object(view, 'get_instance', new=get_instance):

                # get the form page
                response = view.edit()
                self.assertIsInstance(response, Response)
                self.assertEqual(response.status_code, 200)
                self.assertIn('frazzle', response.text)
                # nb. no error
                self.assertNotIn('Required', response.text)

                def persist(setting):
                    self.app.save_setting(self.session, 'foo.bar', setting['value'])
                    self.session.commit()

                # post request to save settings
                self.request.method = 'POST'
                self.request.POST = {
                    'name': 'foo.bar',
                    'value': 'froogle',
                }
                with patch.object(view, 'persist', new=persist):
                    response = view.edit()
                # nb. should get redirect back to view page
                self.assertEqual(response.status_code, 302)
                # setting should be updated in DB
                self.assertEqual(self.app.get_setting(self.session, 'foo.bar'), 'froogle')

                # try another post with invalid data (value is required)
                self.request.method = 'POST'
                self.request.POST = {}
                with patch.object(view, 'persist', new=persist):
                    response = view.edit()
                # nb. should get a form with errors
                self.assertEqual(response.status_code, 200)
                self.assertIn('Required', response.text)
                # setting did not change in DB
                self.assertEqual(self.app.get_setting(self.session, 'foo.bar'), 'froogle')

    def test_delete(self):
        self.pyramid_config.include('wuttaweb.views.common')
        self.pyramid_config.include('wuttaweb.views.auth')
        self.pyramid_config.add_route('settings.create', '/settings/new')
        self.pyramid_config.add_route('settings.view', '/settings/{name}')
        self.pyramid_config.add_route('settings.edit', '/settings/{name}/edit')
        model = self.app.model
        self.app.save_setting(self.session, 'foo.bar', 'frazzle')
        self.session.commit()
        self.assertEqual(self.session.query(model.Setting).count(), 1)

        def get_instance():
            setting = self.session.get(model.Setting, 'foo.bar')
            return {
                'name': setting.name,
                'value': setting.value,
            }

        # sanity/coverage check using /settings/XXX/delete
        self.request.matchdict = {'name': 'foo.bar'}
        with patch.multiple(mod.MasterView, create=True,
                            model_name='Setting',
                            model_key='name',
                            get_index_url=MagicMock(return_value='/settings/'),
                            form_fields=['name', 'value']):
            view = mod.MasterView(self.request)
            with patch.object(view, 'get_instance', new=get_instance):

                # get the form page
                response = view.delete()
                self.assertIsInstance(response, Response)
                self.assertEqual(response.status_code, 200)
                self.assertIn('frazzle', response.text)

                def delete_instance(setting):
                    self.app.delete_setting(self.session, setting['name'])

                self.request.method = 'POST'
                self.request.POST = {}
                with patch.object(view, 'delete_instance', new=delete_instance):

                    # enforces "instance not deletable" rules
                    with patch.object(view, 'is_deletable', return_value=False):
                        response = view.delete()
                    # nb. should get redirect back to view page
                    self.assertEqual(response.status_code, 302)
                    # setting remains in DB
                    self.assertEqual(self.session.query(model.Setting).count(), 1)

                    # post request to delete setting
                    response = view.delete()
                    # nb. should get redirect back to view page
                    self.assertEqual(response.status_code, 302)
                    # setting should be gone from DB
                    self.assertEqual(self.session.query(model.Setting).count(), 0)

    def test_delete_instance(self):
        model = self.app.model
        self.app.save_setting(self.session, 'foo.bar', 'frazzle')
        self.session.commit()
        setting = self.session.query(model.Setting).one()

        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting,
                            form_fields=['name', 'value']):
            view = mod.MasterView(self.request)
            view.delete_instance(setting)
            self.session.commit()
            self.assertEqual(self.session.query(model.Setting).count(), 0)

    def test_delete_bulk(self):
        self.pyramid_config.add_route('settings', '/settings/')
        self.pyramid_config.add_route('progress', '/progress/{key}')
        model = self.app.model
        sample_data = [
            {'name': 'foo1', 'value': 'ONE'},
            {'name': 'foo2', 'value': 'two'},
            {'name': 'foo3', 'value': 'three'},
            {'name': 'foo4', 'value': 'four'},
            {'name': 'foo5', 'value': 'five'},
            {'name': 'foo6', 'value': 'six'},
            {'name': 'foo7', 'value': 'seven'},
            {'name': 'foo8', 'value': 'eight'},
            {'name': 'foo9', 'value': 'nine'},
        ]
        for setting in sample_data:
            self.app.save_setting(self.session, setting['name'], setting['value'])
        self.session.commit()
        sample_query = self.session.query(model.Setting)

        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting):
            view = self.make_view()

            # sanity check on sample data
            grid = view.make_model_grid(session=self.session)
            data = grid.get_visible_data()
            self.assertEqual(len(data), 9)

            # and then let's filter it a little
            self.request.GET = {'value': 's', 'value.verb': 'contains'}
            grid = view.make_model_grid(session=self.session)
            self.assertEqual(len(grid.filters), 2)
            self.assertEqual(len(grid.active_filters), 1)
            data = grid.get_visible_data()
            self.assertEqual(len(data), 2)

            # okay now let's delete those via quick method
            # (user should be redirected back to index)
            with patch.multiple(view,
                                deletable_bulk_quick=True,
                                make_model_grid=MagicMock(return_value=grid)):
                response = view.delete_bulk()
            self.assertEqual(response.status_code, 302)
            self.assertEqual(self.session.query(model.Setting).count(), 7)

            # now use another filter since those records are gone
            self.request.GET = {'name': 'foo2', 'name.verb': 'equal'}
            grid = view.make_model_grid(session=self.session)
            self.assertEqual(len(grid.filters), 2)
            self.assertEqual(len(grid.active_filters), 1)
            data = grid.get_visible_data()
            self.assertEqual(len(data), 1)

            # this time we delete "slowly" with progress
            self.request.session.id = 'ignorethis'
            with patch.multiple(view,
                                deletable_bulk_quick=False,
                                make_model_grid=MagicMock(return_value=grid)):
                with patch.object(mod, 'threading') as threading:
                    response = view.delete_bulk()
                    threading.Thread.return_value.start.assert_called_once_with()
            # nb. user is shown progress page
            self.assertEqual(response.status_code, 200)

    def test_delete_bulk_action(self):
        self.pyramid_config.add_route('settings', '/settings/')
        model = self.app.model
        sample_data = [
            {'name': 'foo1', 'value': 'ONE'},
            {'name': 'foo2', 'value': 'two'},
            {'name': 'foo3', 'value': 'three'},
            {'name': 'foo4', 'value': 'four'},
            {'name': 'foo5', 'value': 'five'},
            {'name': 'foo6', 'value': 'six'},
            {'name': 'foo7', 'value': 'seven'},
            {'name': 'foo8', 'value': 'eight'},
            {'name': 'foo9', 'value': 'nine'},
        ]
        for setting in sample_data:
            self.app.save_setting(self.session, setting['name'], setting['value'])
        self.session.commit()
        sample_query = self.session.query(model.Setting)

        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting):
            view = self.make_view()

            # basic bulk delete
            self.assertEqual(self.session.query(model.Setting).count(), 9)
            settings = self.session.query(model.Setting)\
                                   .filter(model.Setting.value.ilike('%s%'))\
                                   .all()
            self.assertEqual(len(settings), 2)
            view.delete_bulk_action(settings)
            self.session.commit()
            self.assertEqual(self.session.query(model.Setting).count(), 7)

    def test_delete_bulk_thread(self):
        self.pyramid_config.add_route('settings', '/settings/')
        model = self.app.model
        sample_data = [
            {'name': 'foo1', 'value': 'ONE'},
            {'name': 'foo2', 'value': 'two'},
            {'name': 'foo3', 'value': 'three'},
            {'name': 'foo4', 'value': 'four'},
            {'name': 'foo5', 'value': 'five'},
            {'name': 'foo6', 'value': 'six'},
            {'name': 'foo7', 'value': 'seven'},
            {'name': 'foo8', 'value': 'eight'},
            {'name': 'foo9', 'value': 'nine'},
        ]
        for setting in sample_data:
            self.app.save_setting(self.session, setting['name'], setting['value'])
        self.session.commit()
        sample_query = self.session.query(model.Setting)

        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting):
            view = self.make_view()

            # basic delete, no progress
            self.assertEqual(self.session.query(model.Setting).count(), 9)
            settings = self.session.query(model.Setting)\
                                   .filter(model.Setting.value.ilike('%s%'))
            self.assertEqual(settings.count(), 2)
            with patch.object(self.app, 'make_session', return_value=self.session):
                view.delete_bulk_thread(settings)
            self.assertEqual(self.session.query(model.Setting).count(), 7)

            # basic delete, with progress
            settings = self.session.query(model.Setting)\
                                   .filter(model.Setting.name == 'foo1')
            self.assertEqual(settings.count(), 1)
            with patch.object(self.app, 'make_session', return_value=self.session):
                view.delete_bulk_thread(settings, progress=MagicMock())
            self.assertEqual(self.session.query(model.Setting).count(), 6)

            # error, no progress
            settings = self.session.query(model.Setting)\
                                   .filter(model.Setting.name == 'foo2')
            self.assertEqual(settings.count(), 1)
            with patch.object(self.app, 'make_session', return_value=self.session):
                with patch.object(view, 'delete_bulk_action', side_effect=RuntimeError):
                    view.delete_bulk_thread(settings)
            # nb. nothing was deleted
            self.assertEqual(self.session.query(model.Setting).count(), 6)

            # error, with progress
            self.assertEqual(settings.count(), 1)
            with patch.object(self.app, 'make_session', return_value=self.session):
                with patch.object(view, 'delete_bulk_action', side_effect=RuntimeError):
                    view.delete_bulk_thread(settings, progress=MagicMock())
            # nb. nothing was deleted
            self.assertEqual(self.session.query(model.Setting).count(), 6)

    def test_autocomplete(self):
        model = self.app.model

        person1 = model.Person(full_name="George Jones")
        self.session.add(person1)
        person2 = model.Person(full_name="George Strait")
        self.session.add(person2)
        self.session.commit()

        # no results for empty term
        self.request.GET = {}
        view = self.make_view()
        results = view.autocomplete()
        self.assertEqual(len(results), 0)

        # search yields no results
        self.request.GET = {'term': 'sally'}
        view = self.make_view()
        with patch.object(view, 'autocomplete_data', return_value=[]):
            view = self.make_view()
            results = view.autocomplete()
            self.assertEqual(len(results), 0)

        # search yields 2 results
        self.request.GET = {'term': 'george'}
        view = self.make_view()
        with patch.object(view, 'autocomplete_data', return_value=[person1, person2]):
            results = view.autocomplete()
            self.assertEqual(len(results), 2)
            self.assertEqual([res['value'] for res in results],
                             [p.uuid for p in [person1, person2]])

    def test_autocomplete_normalize(self):
        model = self.app.model
        view = self.make_view()

        person = model.Person(full_name="Betty Boop", uuid='bogus')
        normal = view.autocomplete_normalize(person)
        self.assertEqual(normal, {'value': 'bogus',
                                  'label': "Betty Boop"})

    def test_download(self):
        model = self.app.model
        self.app.save_setting(self.session, 'foo', 'bar')
        self.session.commit()

        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting,
                            model_key='name',
                            Session=MagicMock(return_value=self.session)):
            view = self.make_view()
            self.request.matchdict = {'name': 'foo'}

            # 404 if no filename
            response = view.download()
            self.assertEqual(response.status_code, 404)

            # 404 if bad filename
            self.request.GET = {'filename': 'doesnotexist'}
            response = view.download()
            self.assertEqual(response.status_code, 404)

            # 200 if good filename
            foofile = self.write_file('foo.txt', 'foo')
            with patch.object(view, 'download_path', return_value=foofile):
                response = view.download()
                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.content_disposition, 'attachment; filename="foo.txt"')

    def test_execute(self):
        self.pyramid_config.add_route('settings.view', '/settings/{name}')
        self.pyramid_config.add_route('progress', '/progress/{key}')
        model = self.app.model
        self.app.save_setting(self.session, 'foo', 'bar')
        user = model.User(username='barney')
        self.session.add(user)
        self.session.commit()

        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Setting,
                            model_key='name',
                            Session=MagicMock(return_value=self.session)):
            view = self.make_view()
            self.request.matchdict = {'name': 'foo'}
            self.request.session.id = 'mockid'
            self.request.user = user

            # basic usage; user is shown progress page
            with patch.object(mod, 'threading') as threading:
                response = view.execute()
                threading.Thread.return_value.start.assert_called_once_with()
                self.assertEqual(response.status_code, 200)

    def test_execute_thread(self):
        model = self.app.model
        enum = self.app.enum
        user = model.User(username='barney')
        self.session.add(user)
        upgrade = model.Upgrade(description='test', created_by=user,
                                status=enum.UpgradeStatus.PENDING)
        self.session.add(upgrade)
        self.session.commit()

        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Upgrade):
            view = self.make_view()

            # basic execute, no progress
            with patch.object(view, 'execute_instance') as execute_instance:
                view.execute_thread({'uuid': upgrade.uuid}, user.uuid)
                execute_instance.assert_called_once()

            # basic execute, with progress
            with patch.object(view, 'execute_instance') as execute_instance:
                progress = MagicMock()
                view.execute_thread({'uuid': upgrade.uuid}, user.uuid, progress=progress)
                execute_instance.assert_called_once()
                progress.handle_success.assert_called_once_with()

            # error, no progress
            with patch.object(view, 'execute_instance') as execute_instance:
                execute_instance.side_effect = RuntimeError
                view.execute_thread({'uuid': upgrade.uuid}, user.uuid)
                execute_instance.assert_called_once()

            # error, with progress
            with patch.object(view, 'execute_instance') as execute_instance:
                progress = MagicMock()
                execute_instance.side_effect = RuntimeError
                view.execute_thread({'uuid': upgrade.uuid}, user.uuid, progress=progress)
                execute_instance.assert_called_once()
                progress.handle_error.assert_called_once()

    def test_configure(self):
        self.pyramid_config.include('wuttaweb.views.common')
        self.pyramid_config.include('wuttaweb.views.auth')
        model = self.app.model

        # mock settings
        settings = [
            {'name': 'wutta.app_title'},
            {'name': 'wutta.foo', 'value': 'bar'},
            {'name': 'wutta.flag', 'type': bool},
            {'name': 'wutta.number', 'type': int, 'default': 42},
            {'name': 'wutta.value1', 'save_if_empty': True},
            {'name': 'wutta.value2', 'save_if_empty': False},
        ]

        view = mod.MasterView(self.request)
        with patch.object(self.request, 'current_route_url', return_value='/appinfo/configure'):
            with patch.object(mod, 'Session', return_value=self.session):
                with patch.multiple(mod.MasterView, create=True,
                                    model_name='AppInfo',
                                    route_prefix='appinfo',
                                    template_prefix='/appinfo',
                                    creatable=False,
                                    get_index_url=MagicMock(return_value='/appinfo/'),
                                    configure_get_simple_settings=MagicMock(return_value=settings)):

                    # nb. appinfo/configure template requires menu_handlers
                    original_context = view.configure_get_context
                    def get_context(**kw):
                        kw = original_context(**kw)
                        kw['menu_handlers'] = []
                        return kw
                    with patch.object(view, 'configure_get_context', new=get_context):

                        # get the form page
                        response = view.configure(session=self.session)
                        self.assertIsInstance(response, Response)

                        # post request to save settings
                        self.request.method = 'POST'
                        self.request.POST = {
                            'wutta.app_title': 'Wutta',
                            'wutta.foo': 'bar',
                            'wutta.flag': 'true',
                        }
                        response = view.configure(session=self.session)
                        # nb. should get redirect back to configure page
                        self.assertEqual(response.status_code, 302)

                        # should now have 5 settings
                        count = self.session.query(model.Setting).count()
                        self.assertEqual(count, 5)
                        get_setting = functools.partial(self.app.get_setting, self.session)
                        self.assertEqual(get_setting('wutta.app_title'), 'Wutta')
                        self.assertEqual(get_setting('wutta.foo'), 'bar')
                        self.assertEqual(get_setting('wutta.flag'), 'true')
                        self.assertEqual(get_setting('wutta.number'), '42')
                        self.assertEqual(get_setting('wutta.value1'), '')
                        self.assertEqual(get_setting('wutta.value2'), None)

                        # post request to remove settings
                        self.request.method = 'POST'
                        self.request.POST = {'remove_settings': '1'}
                        response = view.configure(session=self.session)
                        # nb. should get redirect back to configure page
                        self.assertEqual(response.status_code, 302)

                        # should now have 0 settings
                        count = self.session.query(model.Setting).count()
                        self.assertEqual(count, 0)

    def test_configure_gather_settings(self):
        view = self.make_view()

        simple_settings = [
            {'name': 'wutta.app_title'},
            {'name': 'wutta.foo'},
            {'name': 'wutta.flag', 'type': bool, 'default': True},
            {'name': 'wutta.number', 'type': int, 'default': 42},
            {'name': 'wutta.value1', 'save_if_empty': True},
            {'name': 'wutta.value2', 'save_if_empty': False},
            {'name': 'wutta.value3', 'save_if_empty': False, 'default': 'baz'},
        ]

        data = {
            'wutta.app_title': 'Poser',
            'wutta.foo': 'bar',
            'wutta.number': 44,
            'wutta.value1': None,
        }

        with patch.object(view, 'configure_get_simple_settings', return_value=simple_settings):
            settings = view.configure_gather_settings(data)
            self.assertEqual(len(settings), 6)
            self.assertEqual(settings, [
                {'name': 'wutta.app_title', 'value': 'Poser'},
                {'name': 'wutta.foo', 'value': 'bar'},
                {'name': 'wutta.flag', 'value': 'false'},
                {'name': 'wutta.number', 'value': '44'},
                {'name': 'wutta.value1', 'value': ''},
                {'name': 'wutta.value3', 'value': 'baz'},
            ])

    ##############################
    # row methods
    ##############################

    def test_collect_row_labels(self):

        # default labels
        view = self.make_view()
        labels = view.collect_row_labels()
        self.assertEqual(labels, {})

        # labels come from all classes; subclass wins
        with patch.object(View, 'row_labels', create=True, new={'foo': "Foo", 'bar': "Bar"}):
            with patch.object(mod.MasterView, 'row_labels', create=True, new={'foo': "FOO FIGHTERS"}):
                view = self.make_view()
                labels = view.collect_row_labels()
                self.assertEqual(labels, {'foo': "FOO FIGHTERS", 'bar': "Bar"})

    def test_set_row_labels(self):
        model = self.app.model
        person = model.Person(full_name="Fred Flintstone")
        self.session.add(person)

        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Person,
                            has_rows=True,
                            row_model_class=model.User):

            # no labels by default
            view = self.make_view()
            grid = view.make_row_model_grid(person, key='person.users', data=[])
            view.set_row_labels(grid)
            self.assertEqual(grid.labels, {})

            # labels come from all classes; subclass wins
            with patch.object(View, 'row_labels', create=True, new={'username': "USERNAME"}):
                with patch.object(mod.MasterView, 'row_labels', create=True, new={'username': "UserName"}):
                    view = self.make_view()
                    grid = view.make_row_model_grid(person, key='person.users', data=[])
                    view.set_row_labels(grid)
                    self.assertEqual(grid.labels, {'username': "UserName"})

    def test_get_row_grid_data(self):
        model = self.app.model
        person = model.Person(full_name="Fred Flintstone")
        self.session.add(person)
        view = self.make_view()
        self.assertRaises(NotImplementedError, view.get_row_grid_data, person)

    def test_get_row_grid_columns(self):

        # no default
        view = self.make_view()
        self.assertIsNone(view.get_row_grid_columns())

        # class may specify
        with patch.object(view, 'row_grid_columns', create=True, new=['foo', 'bar']):
            self.assertEqual(view.get_row_grid_columns(), ['foo', 'bar'])

    def test_get_row_grid_key(self):
        view = self.make_view()
        with patch.multiple(mod.MasterView, create=True,
                            model_key='id',
                            grid_key='widgets'):

            self.request.matchdict = {'id': 42}
            self.assertEqual(view.get_row_grid_key(), 'widgets.42')

    def test_make_row_model_grid(self):
        model = self.app.model
        person = model.Person(full_name="Barney Rubble")
        self.session.add(person)
        self.session.commit()

        self.request.matchdict = {'uuid': person.uuid}
        with patch.multiple(mod.MasterView, create=True,
                            model_class=model.Person):
            view = self.make_view()

            # specify data
            grid = view.make_row_model_grid(person, data=[])
            self.assertIsNone(grid.model_class)
            self.assertEqual(grid.data, [])

            # fetch data
            with patch.object(view, 'get_row_grid_data', return_value=[]):
                grid = view.make_row_model_grid(person)
                self.assertIsNone(grid.model_class)
                self.assertEqual(grid.data, [])

            # view action
            with patch.object(view, 'rows_viewable', new=True):
                with patch.object(view, 'get_row_action_url_view', return_value='#'):
                    grid = view.make_row_model_grid(person, data=[])
                    self.assertEqual(len(grid.actions), 1)
                    self.assertEqual(grid.actions[0].key, 'view')

    def test_get_row_action_url_view(self):
        view = self.make_view()
        row = MagicMock()
        self.assertRaises(NotImplementedError, view.get_row_action_url_view, row, 0)

    def test_get_rows_title(self):
        view = self.make_view()

        # no default
        self.assertIsNone(view.get_rows_title())

        # class may specify
        with patch.object(view, 'rows_title', create=True, new="Mock Rows"):
            self.assertEqual(view.get_rows_title(), "Mock Rows")



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Form widgets

This module defines some custom widgets for use with WuttaWeb.

However for convenience it also makes other Deform widgets available
in the namespace:

* :class:`deform:deform.widget.Widget` (base class)
* :class:`deform:deform.widget.TextInputWidget`
* :class:`deform:deform.widget.TextAreaWidget`
* :class:`deform:deform.widget.PasswordWidget`
* :class:`deform:deform.widget.CheckedPasswordWidget`
* :class:`deform:deform.widget.CheckboxWidget`
* :class:`deform:deform.widget.SelectWidget`
* :class:`deform:deform.widget.CheckboxChoiceWidget`
* :class:`deform:deform.widget.DateInputWidget`
* :class:`deform:deform.widget.DateTimeInputWidget`
* :class:`deform:deform.widget.MoneyInputWidget`
"""

import datetime
import decimal
import os

import colander
import humanize
from deform.widget import (Widget, TextInputWidget, TextAreaWidget,
                           PasswordWidget, CheckedPasswordWidget,
                           CheckboxWidget, SelectWidget, CheckboxChoiceWidget,
                           DateInputWidget, DateTimeInputWidget, MoneyInputWidget)
from webhelpers2.html import HTML

from wuttjamaican.conf import parse_list

from wuttaweb.db import Session
from wuttaweb.grids import Grid


class ObjectRefWidget(SelectWidget):
    """
    Widget for use with model "object reference" fields, e.g.  foreign
    key UUID => TargetModel instance.

    While you may create instances of this widget directly, it
    normally happens automatically when schema nodes of the
    :class:`~wuttaweb.forms.schema.ObjectRef` (sub)type are part of
    the form schema; via
    :meth:`~wuttaweb.forms.schema.ObjectRef.widget_maker()`.

    In readonly mode, this renders a ``<span>`` tag around the
    :attr:`model_instance` (converted to string).

    Otherwise it renders a select (dropdown) element allowing user to
    choose from available records.

    This is a subclass of :class:`deform:deform.widget.SelectWidget`
    and uses these Deform templates:

    * ``select``
    * ``readonly/objectref``

    .. attribute:: model_instance

       Reference to the model record instance, i.e. the "far side" of
       the foreign key relationship.

       .. note::

          You do not need to provide the ``model_instance`` when
          constructing the widget.  Rather, it is set automatically
          when the :class:`~wuttaweb.forms.schema.ObjectRef` type
          instance (associated with the node) is serialized.
    """
    readonly_template = 'readonly/objectref'

    def __init__(self, request, url=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.url = url

    def get_template_values(self, field, cstruct, kw):
        """ """
        values = super().get_template_values(field, cstruct, kw)

        # add url, only if rendering readonly
        readonly = kw.get('readonly', self.readonly)
        if readonly:
            if 'url' not in values and self.url and getattr(field.schema, 'model_instance', None):
                values['url'] = self.url(field.schema.model_instance)

        return values


class NotesWidget(TextAreaWidget):
    """
    Widget for use with "notes" fields.

    In readonly mode, this shows the notes with a background to make
    them stand out a bit more.

    Otherwise it effectively shows a ``<textarea>`` input element.

    This is a subclass of :class:`deform:deform.widget.TextAreaWidget`
    and uses these Deform templates:

    * ``textarea``
    * ``readonly/notes``
    """
    readonly_template = 'readonly/notes'


class WuttaCheckboxChoiceWidget(CheckboxChoiceWidget):
    """
    Custom widget for :class:`python:set` fields.

    This is a subclass of
    :class:`deform:deform.widget.CheckboxChoiceWidget`.

    :param request: Current :term:`request` object.

    It uses these Deform templates:

    * ``checkbox_choice``
    * ``readonly/checkbox_choice``
    """

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()


class WuttaDateWidget(DateInputWidget):
    """
    Custom widget for :class:`python:datetime.date` fields.

    The main purpose of this widget is to leverage
    :meth:`~wuttjamaican:wuttjamaican.app.AppHandler.render_date()`
    for the readonly display.

    It is automatically used for SQLAlchemy mapped classes where the
    field maps to a :class:`sqlalchemy:sqlalchemy.types.Date` column.
    For other (non-mapped) date fields, or mapped datetime fields for
    which a date widget is preferred, use
    :meth:`~wuttaweb.forms.base.Form.set_widget()`.

    This is a subclass of
    :class:`deform:deform.widget.DateInputWidget` and uses these
    Deform templates:

    * ``dateinput``
    """

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get('readonly', self.readonly)
        if readonly and cstruct:
            dt = datetime.datetime.fromisoformat(cstruct)
            return self.app.render_date(dt)

        return super().serialize(field, cstruct, **kw)


class WuttaDateTimeWidget(DateTimeInputWidget):
    """
    Custom widget for :class:`python:datetime.datetime` fields.

    The main purpose of this widget is to leverage
    :meth:`~wuttjamaican:wuttjamaican.app.AppHandler.render_datetime()`
    for the readonly display.

    It is automatically used for SQLAlchemy mapped classes where the
    field maps to a :class:`sqlalchemy:sqlalchemy.types.DateTime`
    column.  For other (non-mapped) datetime fields, you may have to
    use it explicitly via
    :meth:`~wuttaweb.forms.base.Form.set_widget()`.

    This is a subclass of
    :class:`deform:deform.widget.DateTimeInputWidget` and uses these
    Deform templates:

    * ``datetimeinput``
    """

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get('readonly', self.readonly)
        if readonly and cstruct:
            dt = datetime.datetime.fromisoformat(cstruct)
            return self.app.render_datetime(dt)

        return super().serialize(field, cstruct, **kw)


class WuttaMoneyInputWidget(MoneyInputWidget):
    """
    Custom widget for "money" fields.  This is used by default for
    :class:`~wuttaweb.forms.schema.WuttaMoney` type nodes.

    The main purpose of this widget is to leverage
    :meth:`~wuttjamaican:wuttjamaican.app.AppHandler.render_currency()`
    for the readonly display.

    This is a subclass of
    :class:`deform:deform.widget.MoneyInputWidget` and uses these
    Deform templates:

    * ``moneyinput``

    :param request: Current :term:`request` object.

    :param scale: If this kwarg is specified, it will be passed along
       to ``render_currency()`` call.
    """

    def __init__(self, request, *args, **kwargs):
        self.scale = kwargs.pop('scale', 2)
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get('readonly', self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag('span')
            cstruct = decimal.Decimal(cstruct)
            text = self.app.render_currency(cstruct, scale=self.scale)
            return HTML.tag('span', c=[text])

        return super().serialize(field, cstruct, **kw)


class FileDownloadWidget(Widget):
    """
    Widget for use with :class:`~wuttaweb.forms.schema.FileDownload`
    fields.

    This only supports readonly, and shows a hyperlink to download the
    file.  Link text is the filename plus file size.

    This is a subclass of :class:`deform:deform.widget.Widget` and
    uses these Deform templates:

    * ``readonly/filedownload``

    :param request: Current :term:`request` object.

    :param url: Optional URL for hyperlink.  If not specified, file
       name/size is shown with no hyperlink.
    """
    readonly_template = 'readonly/filedownload'

    def __init__(self, request, *args, **kwargs):
        self.url = kwargs.pop('url', None)
        super().__init__(*args, **kwargs)
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()

    def serialize(self, field, cstruct, **kw):
        """ """
        # nb. readonly is the only way this rolls
        kw['readonly'] = True
        template = self.readonly_template

        path = cstruct or None
        if path:
            kw.setdefault('filename', os.path.basename(path))
            kw.setdefault('filesize', self.readable_size(path))
            if self.url:
                kw.setdefault('url', self.url)

        else:
            kw.setdefault('filename', None)
            kw.setdefault('filesize', None)

        kw.setdefault('url', None)
        values = self.get_template_values(field, cstruct, kw)
        return field.renderer(template, **values)

    def readable_size(self, path):
        """ """
        try:
            size = os.path.getsize(path)
        except os.error:
            size = 0
        return humanize.naturalsize(size)


class GridWidget(Widget):
    """
    Widget for fields whose data is represented by a :term:`grid`.

    This is a subclass of :class:`deform:deform.widget.Widget` but
    does not use any Deform templates.

    This widget only supports "readonly" mode, is not editable.  It is
    merely a convenience around the grid itself, which does the heavy
    lifting.

    Instead of creating this widget directly you probably should call
    :meth:`~wuttaweb.forms.base.Form.set_grid()` on your form.

    :param request: Current :term:`request` object.

    :param grid: :class:`~wuttaweb.grids.base.Grid` instance, used to
       display the field data.
    """

    def __init__(self, request, grid, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.grid = grid

    def serialize(self, field, cstruct, **kw):
        """
        This widget simply calls
        :meth:`~wuttaweb.grids.base.Grid.render_table_element()` on
        the ``grid`` to serialize.
        """
        readonly = kw.get('readonly', self.readonly)
        if not readonly:
            raise NotImplementedError("edit not allowed for this widget")

        return self.grid.render_table_element()


class RoleRefsWidget(WuttaCheckboxChoiceWidget):
    """
    Widget for use with User
    :attr:`~wuttjamaican:wuttjamaican.db.model.auth.User.roles` field.
    This is the default widget for the
    :class:`~wuttaweb.forms.schema.RoleRefs` type.

    This is a subclass of :class:`WuttaCheckboxChoiceWidget`.
    """
    readonly_template = 'readonly/rolerefs'

    def serialize(self, field, cstruct, **kw):
        """ """
        model = self.app.model

        # special logic when field is editable
        readonly = kw.get('readonly', self.readonly)
        if not readonly:

            # but does not apply if current user is root
            if not self.request.is_root:
                auth = self.app.get_auth_handler()
                admin = auth.get_role_administrator(self.session)

                # prune admin role from values list; it should not be
                # one of the options since current user is not admin
                values = kw.get('values', self.values)
                values = [val for val in values
                          if val[0] != admin.uuid]
                kw['values'] = values

        else: # readonly

            # roles
            roles = []
            if cstruct:
                for uuid in cstruct:
                    role = self.session.get(model.Role, uuid)
                    if role:
                        roles.append(role)
            kw['roles'] = roles

            # url
            url = lambda role: self.request.route_url('roles.view', uuid=role.uuid)
            kw['url'] = url

        # default logic from here
        return super().serialize(field, cstruct, **kw)


class UserRefsWidget(WuttaCheckboxChoiceWidget):
    """
    Widget for use with Role
    :attr:`~wuttjamaican:wuttjamaican.db.model.auth.Role.users` field.
    This is the default widget for the
    :class:`~wuttaweb.forms.schema.UserRefs` type.

    This is a subclass of :class:`WuttaCheckboxChoiceWidget`; however
    it only supports readonly mode and does not use a template.
    Rather, it generates and renders a
    :class:`~wuttaweb.grids.base.Grid` showing the users list.
    """

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get('readonly', self.readonly)
        if not readonly:
            raise NotImplementedError("edit not allowed for this widget")

        model = self.app.model
        columns = ['username', 'active']

        # generate data set for users
        users = []
        if cstruct:
            for uuid in cstruct:
                user = self.session.get(model.User, uuid)
                if user:
                    users.append(dict([(key, getattr(user, key))
                                       for key in columns + ['uuid']]))

        # do not render if no data
        if not users:
            return HTML.tag('span')

        # grid
        grid = Grid(self.request, key='roles.view.users',
                    columns=columns, data=users)

        # view action
        if self.request.has_perm('users.view'):
            url = lambda user, i: self.request.route_url('users.view', uuid=user['uuid'])
            grid.add_action('view', icon='eye', url=url)
            grid.set_link('person')
            grid.set_link('username')

        # edit action
        if self.request.has_perm('users.edit'):
            url = lambda user, i: self.request.route_url('users.edit', uuid=user['uuid'])
            grid.add_action('edit', url=url)

        # render as simple <b-table>
        # nb. must indicate we are a part of this form
        form = getattr(field.parent, 'wutta_form', None)
        return grid.render_table_element(form)


class PermissionsWidget(WuttaCheckboxChoiceWidget):
    """
    Widget for use with Role
    :attr:`~wuttjamaican:wuttjamaican.db.model.auth.Role.permissions`
    field.

    This is a subclass of :class:`WuttaCheckboxChoiceWidget`.  It uses
    these Deform templates:

    * ``permissions``
    * ``readonly/permissions``
    """
    template = 'permissions'
    readonly_template = 'readonly/permissions'

    def serialize(self, field, cstruct, **kw):
        """ """
        kw.setdefault('permissions', self.permissions)

        if 'values' not in kw:
            values = []
            for gkey, group in self.permissions.items():
                for pkey, perm in group['perms'].items():
                    values.append((pkey, perm['label']))
            kw['values'] = values

        return super().serialize(field, cstruct, **kw)


class EmailRecipientsWidget(TextAreaWidget):
    """
    Widget for :term:`email setting` recipient fields (``To``, ``Cc``,
    ``Bcc``).

    This is a subclass of
    :class:`deform:deform.widget.TextAreaWidget`.  It uses these
    Deform templates:

    * ``textarea``
    * ``readonly/email_recips``

    See also the :class:`~wuttaweb.forms.schema.EmailRecipients`
    schema type, which uses this widget.
    """
    readonly_template = 'readonly/email_recips'

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get('readonly', self.readonly)
        if readonly:
            kw['recips'] = parse_list(cstruct or '')

        return super().serialize(field, cstruct, **kw)

    def deserialize(self, field, pstruct):
        """ """
        if pstruct is colander.null:
            return colander.null

        values = [value for value in parse_list(pstruct)
                  if value]
        return ', '.join(values)


class BatchIdWidget(Widget):
    """
    Widget for use with the
    :attr:`~wuttjamaican:wuttjamaican.db.model.batch.BatchMixin.id`
    field of a :term:`batch` model.

    This widget is "always" read-only and renders the Batch ID as
    zero-padded 8-char string
    """

    def serialize(self, field, cstruct, **kw):
        """ """
        if cstruct is colander.null:
            return colander.null

        batch_id = int(cstruct)
        return f'{batch_id:08d}'
